# -*- coding: utf-8 -*-
'''
Manage Elasticache with boto3
=============================

.. versionadded:: 2017.7.0

Create, destroy and update Elasticache clusters. Be aware that this interacts
with Amazon's services, and so may incur charges.

This module uses boto3 behind the scenes - as a result it inherits any limitations
it boto3's implementation of the AWS API.  It is also designed to as directly as
possible leverage boto3's parameter naming and semantics.  This allows one to use
http://boto3.readthedocs.io/en/latest/reference/services/elasticache.html as an
excellent source for details too involved to reiterate here.

Note:  This module is designed to be transparent ("intentionally ignorant" is the
phrase I used to describe it to my boss) to new AWS / boto options - since all
AWS API params are passed directly through both the state and executions modules,
any new args to existing functions which become available after this documentation
is written should work immediately.

Brand new API calls, of course, would still require new functions to be added :)

This module accepts explicit elasticache credentials but can also utilize IAM
roles assigned to the instance through Instance Profiles. Dynamic credentials are
then automatically obtained from AWS API and no further configuration is necessary.
More information is available
`here <http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/iam-roles-for-amazon-ec2.html>`_.

If IAM roles are not used you need to specify them either in a pillar file or
in the minion's config file:

.. code-block:: yaml

    elasticache.keyid: GKTADJGHEIQSXMKKRBJ08H
    elasticache.key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs

It's also possible to specify ``key``, ``keyid`` and ``region`` via a profile, either
passed in as a dict, or as a string to pull from pillars or minion config:

.. code-block:: yaml

    myprofile:
      keyid: GKTADJGHEIQSXMKKRBJ08H
      key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
        region: us-east-1

.. code-block:: yaml

    Ensure myelasticache exists:
      boto3_elasticache.present:
        - name: myelasticache
        - engine: redis
        - cache_node_type: cache.t1.micro
        - num_cache_nodes: 1
        - notification_topic_arn: arn:aws:sns:us-east-1:879879:my-sns-topic
        - region: us-east-1
        - keyid: GKTADJGHEIQSXMKKRBJ08H
        - key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs

.. code-block:: yaml

    # Using a profile from pillars
    Ensure myelasticache exists:
      boto3_elasticache.present:
        - name: myelasticache
        - engine: redis
        - cache_node_type: cache.t1.micro
        - num_cache_nodes: 1
        - notification_topic_arn: arn:aws:sns:us-east-1:879879:my-sns-topic
        - region: us-east-1
        - profile: myprofile

.. code-block:: yaml

    # Passing in a profile
    Ensure myelasticache exists:
      boto3_elasticache.present:
        - name: myelasticache
        - engine: redis
        - cache_node_type: cache.t1.micro
        - num_cache_nodes: 1
        - notification_topic_arn: arn:aws:sns:us-east-1:879879:my-sns-topic
        - region: us-east-1
        - profile:
            keyid: GKTADJGHEIQSXMKKRBJ08H
            key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
'''

# Import Python Libs
from __future__ import absolute_import, print_function, unicode_literals


def __virtual__():
    '''
    Only load if boto is available.
    '''
    if 'boto3_elasticache.cache_cluster_exists' in __salt__:
        return 'boto3_elasticache'
    else:
        return False


def _diff_cache_cluster(current, desired):
    '''
    If you need to enhance what modify_cache_cluster() considers when deciding what is to be
    (or can be) updated, add it to 'modifiable' below.  It's a dict mapping the param as used
    in modify_cache_cluster() to that in describe_cache_clusters().  Any data fiddlery that
    needs to be done to make the mappings meaningful should be done in the munging section
    below as well.

    This function will ONLY touch settings that are explicitly called out in 'desired' - any
    settings which might have previously been changed from their 'default' values will not be
    changed back simply by leaving them out of 'desired'.  This is both intentional, and
    much, much easier to code :)
    '''
    ### The data formats are annoyingly (and as far as I can can tell, unnecessarily)
    ### different - we have to munge to a common format to compare...
    if current.get('SecurityGroups') is not None:
        current['SecurityGroupIds'] = [s['SecurityGroupId'] for s in current['SecurityGroups']]
    if current.get('CacheSecurityGroups') is not None:
        current['CacheSecurityGroupNames'] = [c['CacheSecurityGroupName'] for c in current['CacheSecurityGroups']]
    if current.get('NotificationConfiguration') is not None:
        current['NotificationTopicArn'] = current['NotificationConfiguration']['TopicArn']
        current['NotificationTopicStatus'] = current['NotificationConfiguration']['TopicStatus']
    if current.get('CacheParameterGroup') is not None:
        current['CacheParameterGroupName'] = current['CacheParameterGroup']['CacheParameterGroupName']

    modifiable = {
        'AutoMinorVersionUpgrade': 'AutoMinorVersionUpgrade',
        'AZMode': 'AZMode',
        'CacheNodeType': 'CacheNodeType',
        'CacheNodeIdsToRemove': None,
        'CacheParameterGroupName': 'CacheParameterGroupName',
        'CacheSecurityGroupNames': 'CacheSecurityGroupNames',
        'EngineVersion': 'EngineVersion',
        'NewAvailabilityZones': None,
        'NotificationTopicArn': 'NotificationTopicArn',
        'NotificationTopicStatus': 'NotificationTopicStatus',
        'NumCacheNodes': 'NumCacheNodes',
        'PreferredMaintenanceWindow': 'PreferredMaintenanceWindow',
        'SecurityGroupIds': 'SecurityGroupIds',
        'SnapshotRetentionLimit': 'SnapshotRetentionLimit',
        'SnapshotWindow': 'SnapshotWindow'
    }

    need_update = {}
    for m, o in modifiable.items():
        if m in desired:
            if not o:
                # Always pass these through - let AWS do the math...
                need_update[m] = desired[m]
            else:
                if m in current:
                    # Equivalence testing works fine for current simple type comparisons
                    # This might need enhancement if more complex structures enter the picture
                    if current[m] != desired[m]:
                        need_update[m] = desired[m]
    return need_update


def cache_cluster_present(name, wait=900, security_groups=None, region=None, key=None,
                          keyid=None, profile=None, **args):
    '''
    Ensure a given cache cluster exists.

    name
        Name of the cache cluster (cache cluster id).

    wait
        Integer describing how long, in seconds, to wait for confirmation from AWS that the
        resource is in the desired state.  Zero meaning to return success or failure immediately
        of course.  Note that waiting for the cluster to become available is generally the
        better course, as failure to do so will often lead to subsequent failures when managing
        dependent resources.

    security_groups
        One or more VPC security groups (names and/or IDs) associated with the cache cluster.

        .. note::
            This is additive with any sec groups provided via the
            SecurityGroupIds parameter below.  Use this parameter ONLY when you
            are creating a cluster in a VPC.

    CacheClusterId
        The node group (shard) identifier. This parameter is stored as a lowercase string.

        Constraints:

        - A name must contain from 1 to 20 alphanumeric characters or hyphens.
        - The first character must be a letter.
        - A name cannot end with a hyphen or contain two consecutive hyphens.

        .. note::
            In general this parameter is not needed, as 'name' is used if it's
            not provided.

    ReplicationGroupId
        The ID of the replication group to which this cache cluster should belong. If this
        parameter is specified, the cache cluster is added to the specified replication
        group as a read replica; otherwise, the cache cluster is a standalone primary that
        is not part of any replication group.  If the specified replication group is
        Multi-AZ enabled and the Availability Zone is not specified, the cache cluster is
        created in Availability Zones that provide the best spread of read replicas across
        Availability Zones.

        .. note:
            This parameter is ONLY valid if the Engine parameter is redis. Due
            to current limitations on Redis (cluster mode disabled), this
            parameter is not supported on Redis (cluster mode enabled)
            replication groups.

    AZMode
        Specifies whether the nodes in this Memcached cluster are created in a single
        Availability Zone or created across multiple Availability Zones in the cluster's
        region. If the AZMode and PreferredAvailabilityZones are not specified,
        ElastiCache assumes single-az mode.

        .. note::
            This parameter is ONLY supported for Memcached cache clusters.

    PreferredAvailabilityZone
        The EC2 Availability Zone in which the cache cluster is created.  All nodes
        belonging to this Memcached cache cluster are placed in the preferred Availability
        Zone. If you want to create your nodes across multiple Availability Zones, use
        PreferredAvailabilityZones.

        Default:  System chosen Availability Zone.

    PreferredAvailabilityZones
        A list of the Availability Zones in which cache nodes are created. The order of
        the zones in the list is not important.  The number of Availability Zones listed
        must equal the value of NumCacheNodes.  If you want all the nodes in the same
        Availability Zone, use PreferredAvailabilityZone instead, or repeat the
        Availability Zone multiple times in the list.

        Default:  System chosen Availability Zones.

        .. note::
            This option is ONLY supported on Memcached.

            If you are creating your cache cluster in an Amazon VPC
            (recommended) you can only locate nodes in Availability Zones that
            are associated with the subnets in the selected subnet group.

    NumCacheNodes
        The initial (integer) number of cache nodes that the cache cluster has.

        .. note::
            For clusters running Redis, this value must be 1.

            For clusters running Memcached, this value must be between 1 and 20.

    CacheNodeType
        The compute and memory capacity of the nodes in the node group (shard).
        Valid node types (and pricing for them) are exhaustively described at
        https://aws.amazon.com/elasticache/pricing/

        .. note::
            All T2 instances must be created in a VPC

           Redis backup/restore is not supported for Redis (cluster mode
           disabled) T1 and T2 instances. Backup/restore is supported on Redis
           (cluster mode enabled) T2 instances.

           Redis Append-only files (AOF) functionality is not supported for T1
           or T2 instances.

    Engine
        The name of the cache engine to be used for this cache cluster.  Valid values for
        this parameter are:  memcached | redis

    EngineVersion
        The version number of the cache engine to be used for this cache cluster. To view
        the supported cache engine versions, use the DescribeCacheEngineVersions operation.

        .. note::
            You can upgrade to a newer engine version but you cannot downgrade
            to an earlier engine version. If you want to use an earlier engine
            version, you must delete the existing cache cluster or replication
            group and create it anew with the earlier engine version.

    CacheParameterGroupName
        The name of the parameter group to associate with this cache cluster. If this
        argument is omitted, the default parameter group for the specified engine is used.
        You cannot use any parameter group which has cluster-enabled='yes' when creating
        a cluster.

    CacheSubnetGroupName
        The name of the Cache Subnet Group to be used for the cache cluster.  Use this
        parameter ONLY when you are creating a cache cluster within a VPC.

        .. note::
            If you're going to launch your cluster in an Amazon VPC, you need
            to create a subnet group before you start creating a cluster.

    CacheSecurityGroupNames
        A list of Cache Security Group names to associate with this cache cluster.  Use
        this parameter ONLY when you are creating a cache cluster outside of a VPC.

    SecurityGroupIds
        One or more VPC security groups associated with the cache cluster.  Use this
        parameter ONLY when you are creating a cache cluster within a VPC.

    Tags
        A list of tags to be added to this resource.  Note that due to shortcomings in the
        AWS API for Elasticache, these can only be set during resource creation - later
        modification is not (currently) supported.

    SnapshotArns
        A single-element string list containing an Amazon Resource Name (ARN) that
        uniquely identifies a Redis RDB snapshot file stored in Amazon S3. The snapshot
        file is used to populate the node group (shard). The Amazon S3 object name in
        the ARN cannot contain any commas.

        .. note::
            This parameter is ONLY valid if the Engine parameter is redis.

    SnapshotName
        The name of a Redis snapshot from which to restore data into the new node group
        (shard). The snapshot status changes to restoring while the new node group (shard)
        is being created.

        .. note::
            This parameter is ONLY valid if the Engine parameter is redis.

    PreferredMaintenanceWindow
        Specifies the weekly time range during which maintenance on the cache cluster is
        permitted.  It is specified as a range in the format ddd:hh24:mi-ddd:hh24:mi
        (24H Clock UTC).  The minimum maintenance window is a 60 minute period.
        Valid values for ddd are:  sun, mon, tue, wed, thu, fri, sat

        Example:  sun:23:00-mon:01:30

    Port
        The port number on which each of the cache nodes accepts connections.

        Default:  6379

    NotificationTopicArn
        The Amazon Resource Name (ARN) of the Amazon Simple Notification Service (SNS)
        topic to which notifications are sent.

        .. note::
            The Amazon SNS topic owner must be the same as the cache cluster
            owner.

    AutoMinorVersionUpgrade
        This (boolean) parameter is currently disabled.

    SnapshotRetentionLimit
        The number of days for which ElastiCache retains automatic snapshots before
        deleting them.

        Default:  0 (i.e., automatic backups are disabled for this cache cluster).

        .. note::
            This parameter is ONLY valid if the Engine parameter is redis.

    SnapshotWindow
        The daily time range (in UTC) during which ElastiCache begins taking a daily
        snapshot of your node group (shard).  If you do not specify this parameter,
        ElastiCache automatically chooses an appropriate time range.

        Example:  05:00-09:00

        .. note::
            This parameter is ONLY valid if the Engine parameter is redis.

    AuthToken
        The password used to access a password protected server.

        Password constraints:

        - Must be only printable ASCII characters.
        - Must be at least 16 characters and no more than 128 characters in length.
        - Cannot contain any of the following characters: '/', '"', or "@".

    CacheNodeIdsToRemove
        A list of cache node IDs to be removed. A node ID is a numeric identifier (0001, 0002,
        etc.).  This parameter is only valid when NumCacheNodes is less than the existing number of
        cache nodes.  The number of cache node IDs supplied in this parameter must match the
        difference between the existing number of cache nodes in the cluster or pending cache nodes,
        whichever is greater, and the value of NumCacheNodes in the request.

    NewAvailabilityZones
        The list of Availability Zones where the new Memcached cache nodes are created.
        This parameter is only valid when NumCacheNodes in the request is greater than the sum of
        the number of active cache nodes and the number of cache nodes pending creation (which may
        be zero).  The number of Availability Zones supplied in this list must match the cache nodes
        being added in this request.
        Note:  This option is only supported on Memcached clusters.

    NotificationTopicStatus
        The status of the SNS notification topic.  Notifications are sent only if the status is active.

        Valid values:  active | inactive

    region
        Region to connect to.

    key
        Secret key to be used.

    keyid
        Access key to be used.

    profile
        A dict with region, key and keyid, or a pillar key (string) that
        contains a dict with region, key and keyid.
    '''
    ret = {'name': name, 'result': True, 'comment': '', 'changes': {}}
    args = dict([(k, v) for k, v in args.items() if not k.startswith('_')])
    current = __salt__['boto3_elasticache.'
                       'describe_cache_clusters'](name, region=region, key=key,
                                                  keyid=keyid, profile=profile)
    if current:
        check_update = True
    else:
        check_update = False
        only_on_modify = [
            'CacheNodeIdsToRemove',
            'NewAvailabilityZones',
            'NotificationTopicStatus'
        ]
        create_args = {}
        for k, v in args.items():
            if k in only_on_modify:
                check_update = True
            else:
                create_args[k] = v
        if __opts__['test']:
            ret['comment'] = 'Cache cluster {0} would be created.'.format(name)
            ret['result'] = None
            return ret
        created = __salt__['boto3_elasticache.'
                           'create_cache_cluster'](name, wait=wait, security_groups=security_groups,
                                                   region=region, key=key, keyid=keyid,
                                                   profile=profile, **create_args)
        if created:
            new = __salt__['boto3_elasticache.'
                           'describe_cache_clusters'](name, region=region, key=key,
                                                      keyid=keyid, profile=profile)
            ret['comment'] = 'Cache cluster {0} was created.'.format(name)
            ret['changes']['old'] = None
            ret['changes']['new'] = new[0]
        else:
            ret['result'] = False
            ret['comment'] = 'Failed to create {0} cache cluster.'.format(name)

    if check_update:
        # Refresh this in case we're updating from 'only_on_modify' above...
        updated = __salt__['boto3_elasticache.'
                           'describe_cache_clusters'](name, region=region, key=key,
                                                      keyid=keyid, profile=profile)
        need_update = _diff_cache_cluster(updated['CacheClusters'][0], args)
        if need_update:
            if __opts__['test']:
                ret['comment'] = 'Cache cluster {0} would be modified.'.format(name)
                ret['result'] = None
                return ret
            modified = __salt__['boto3_elasticache.'
                                'modify_cache_cluster'](name, wait=wait,
                                                        security_groups=security_groups,
                                                        region=region, key=key, keyid=keyid,
                                                        profile=profile, **need_update)
            if modified:
                new = __salt__['boto3_elasticache.'
                               'describe_cache_clusters'](name, region=region, key=key,
                                                          keyid=keyid, profile=profile)
                if ret['comment']:  # 'create' just ran...
                    ret['comment'] += ' ... and then immediately modified.'
                else:
                    ret['comment'] = 'Cache cluster {0} was modified.'.format(name)
                    ret['changes']['old'] = current
                ret['changes']['new'] = new[0]
            else:
                ret['result'] = False
                ret['comment'] = 'Failed to modify cache cluster {0}.'.format(name)
        else:
            ret['comment'] = 'Cache cluster {0} is in the desired state.'.format(name)
    return ret


def cache_cluster_absent(name, wait=600, region=None, key=None, keyid=None,
                         profile=None, **args):
    '''
    Ensure a given cache cluster is deleted.

    name
        Name of the cache cluster.

    wait
        Integer describing how long, in seconds, to wait for confirmation from AWS that the
        resource is in the desired state.  Zero meaning to return success or failure immediately
        of course.  Note that waiting for the cluster to become available is generally the
        better course, as failure to do so will often lead to subsequent failures when managing
        dependent resources.

    CacheClusterId
        The node group (shard) identifier.
        Note:  In general this parameter is not needed, as 'name' is used if it's not provided.

    FinalSnapshotIdentifier
        The user-supplied name of a final cache cluster snapshot.  This is the unique name
        that identifies the snapshot.  ElastiCache creates the snapshot, and then deletes the
        cache cluster immediately afterward.

    region
        Region to connect to.

    key
        Secret key to be used.

    keyid
        Access key to be used.

    profile
        A dict with region, key and keyid, or a pillar key (string)
        that contains a dict with region, key and keyid.
    '''
    ret = {'name': name, 'result': True, 'comment': '', 'changes': {}}
    args = dict([(k, v) for k, v in args.items() if not k.startswith('_')])
    exists = __salt__['boto3_elasticache.'
                      'cache_cluster_exists'](name, region=region, key=key,
                                              keyid=keyid, profile=profile)
    if exists:
        if __opts__['test']:
            ret['comment'] = 'Cache cluster {0} would be removed.'.format(name)
            ret['result'] = None
            return ret
        deleted = __salt__['boto3_elasticache.'
                           'delete_cache_cluster'](name, wait=wait, region=region, key=key,
                                                   keyid=keyid, profile=profile, **args)
        if deleted:
            ret['changes']['old'] = name
            ret['changes']['new'] = None
        else:
            ret['result'] = False
            ret['comment'] = 'Failed to delete {0} cache cluster.'.format(name)
    else:
        ret['comment'] = 'Cache cluster {0} already absent.'.format(name)
    return ret


def _diff_replication_group(current, desired):
    '''
    If you need to enhance what modify_replication_group() considers when deciding what is to be
    (or can be) updated, add it to 'modifiable' below.  It's a dict mapping the param as used
    in modify_replication_group() to that in describe_replication_groups().  Any data fiddlery
    that needs to be done to make the mappings meaningful should be done in the munging section
    below as well.

    This function will ONLY touch settings that are explicitly called out in 'desired' - any
    settings which might have previously been changed from their 'default' values will not be
    changed back simply by leaving them out of 'desired'.  This is both intentional, and
    much, much easier to code :)
    '''
    if current.get('AutomaticFailover') is not None:
        current['AutomaticFailoverEnabled'] = True if current['AutomaticFailover'] in ('enabled',
                'enabling') else False

    modifiable = {
        # Amazingly, the AWS API provides NO WAY to query the current state of most repl group
        # settings!  All we can do is send a modify op with the desired value, just in case it's
        # different.  And THEN, we can't determine if it's been changed!  Stupid?  YOU BET!
        'AutomaticFailoverEnabled': 'AutomaticFailoverEnabled',
        'AutoMinorVersionUpgrade': None,
        'CacheNodeType': None,
        'CacheParameterGroupName': None,
        'CacheSecurityGroupNames': None,
        'EngineVersion': None,
        'NotificationTopicArn': None,
        'NotificationTopicStatus': None,
        'PreferredMaintenanceWindow': None,
        'PrimaryClusterId': None,
        'ReplicationGroupDescription': 'Description',
        'SecurityGroupIds': None,
        'SnapshotRetentionLimit': 'SnapshotRetentionLimit',
        'SnapshottingClusterId': 'SnapshottingClusterId',
        'SnapshotWindow': 'SnapshotWindow'
    }

    need_update = {}
    for m, o in modifiable.items():
        if m in desired:
            if not o:
                # Always pass these through - let AWS do the math...
                need_update[m] = desired[m]
            else:
                if m in current:
                    # Equivalence testing works fine for current simple type comparisons
                    # This might need enhancement if more complex structures enter the picture
                    if current[m] != desired[m]:
                        need_update[m] = desired[m]
    return need_update


def replication_group_present(name, wait=900, security_groups=None, region=None, key=None,
                              keyid=None, profile=None, **args):
    '''
    Ensure a replication group exists and is in the given state.

    name
        Name of replication group

    wait
        Integer describing how long, in seconds, to wait for confirmation from AWS that the
        resource is in the desired state.  Zero meaning to return success or failure immediately
        of course.  Note that waiting for the cluster to become available is generally the
        better course, as failure to do so will often lead to subsequent failures when managing
        dependent resources.

    security_groups
        One or more VPC security groups (names and/or IDs) associated with the cache cluster.

        .. note::
            This is additive with any sec groups provided via the
            SecurityGroupIds parameter below.  Use this parameter ONLY when you
            are creating a cluster in a VPC.

    ReplicationGroupId
        The replication group identifier. This parameter is stored as a lowercase string.

        Constraints:

        - A name must contain from 1 to 20 alphanumeric characters or hyphens.
        - The first character must be a letter.
        - A name cannot end with a hyphen or contain two consecutive hyphens.

        .. note::
            In general this parameter is not needed, as 'name' is used if it's
            not provided.

    ReplicationGroupDescription
        A user-created description for the replication group.

    PrimaryClusterId
        The identifier of the cache cluster that serves as the primary for this replication group.
        This cache cluster must already exist and have a status of available.  This parameter is
        not required if NumCacheClusters, NumNodeGroups, or ReplicasPerNodeGroup is specified.

    AutomaticFailoverEnabled
        Specifies whether a read-only replica is automatically promoted to read/write primary if
        the existing primary fails.  If true, Multi-AZ is enabled for this replication group. If
        false, Multi-AZ is disabled for this replication group.

        Default:  False

        .. note::
            AutomaticFailoverEnabled must be enabled for Redis (cluster mode
            enabled) replication groups.

            ElastiCache Multi-AZ replication groups is not supported on:

            - Redis versions earlier than 2.8.6.
            - Redis (cluster mode disabled): T1 and T2 node types.
            - Redis (cluster mode enabled): T2 node types.

    NumCacheClusters
        The number of clusters this replication group initially has.  This parameter is not used
        if there is more than one node group (shard). You should use ReplicasPerNodeGroup instead.
        If Multi-AZ is enabled , the value of this parameter must be at least 2.  The maximum
        permitted value for NumCacheClusters is 6 (primary plus 5 replicas).

    PreferredCacheClusterAZs
        A list of EC2 Availability Zones in which the replication group's cache clusters are
        created. The order of the Availability Zones in the list is the order in which clusters
        are allocated. The primary cluster is created in the first AZ in the list.  This parameter
        is not used if there is more than one node group (shard).  You should use
        NodeGroupConfiguration instead.  The number of Availability Zones listed must equal the
        value of NumCacheClusters.

        Default:  System chosen Availability Zones.

        .. note::
            If you are creating your replication group in an Amazon VPC
            (recommended), you can only locate cache clusters in Availability
            Zones associated with the subnets in the selected subnet group.

    NumNodeGroups
        An optional parameter that specifies the number of node groups (shards)
        for this Redis (cluster mode enabled) replication group. For Redis
        (cluster mode disabled) either omit this parameter or set it to 1.

        Default:  1

    ReplicasPerNodeGroup
        An optional parameter that specifies the number of replica nodes in
        each node group (shard). Valid values are:  0 to 5

    NodeGroupConfiguration
        A list of node group (shard) configuration options. Each node group (shard) configuration
        has the following:  Slots, PrimaryAvailabilityZone, ReplicaAvailabilityZones, ReplicaCount.
        If you're creating a Redis (cluster mode disabled) or a Redis (cluster mode enabled)
        replication group, you can use this parameter to configure one node group (shard) or you
        can omit this parameter.  For fiddly details of the expected data layout of this param, see
        http://boto3.readthedocs.io/en/latest/reference/services/elasticache.html?#ElastiCache.Client.create_replication_group

    CacheNodeType
        The compute and memory capacity of the nodes in the node group (shard).
        See https://aws.amazon.com/elasticache/pricing/ for current sizing, prices, and constraints.

        .. note:
            All T2 instances are created in an Amazon Virtual Private Cloud
            (Amazon VPC). Backup/restore is not supported for Redis (cluster
            mode disabled) T1 and T2 instances. Backup/restore is supported on
            Redis (cluster mode enabled) T2 instances. Redis Append-only files
            (AOF) functionality is not supported for T1 or T2 instances.

    Engine
        The name of the cache engine to be used for the cache clusters in this replication group.

    EngineVersion
        The version number of the cache engine to be used for the cache clusters in this replication
        group. To view the supported cache engine versions, use the DescribeCacheEngineVersions
        operation.

        .. note::
            You can upgrade to a newer engine version but you cannot downgrade
            to an earlier engine version. If you want to use an earlier engine
            version, you must delete the existing cache cluster or replication
            group and create it anew with the earlier engine version.

    CacheParameterGroupName
        The name of the parameter group to associate with this replication group. If this argument
        is omitted, the default cache parameter group for the specified engine is used.

        .. note::
            If you are running Redis version 3.2.4 or later, only one node
            group (shard), and want to use a default parameter group, we
            recommend that you specify the parameter group by name.

            To create a Redis (cluster mode disabled) replication group, use
            CacheParameterGroupName=default.redis3.2

            To create a Redis (cluster mode enabled) replication group, use
            CacheParameterGroupName=default.redis3.2.cluster.on

    CacheSubnetGroupName
        The name of the cache subnet group to be used for the replication group.

        .. note::
            If you're going to launch your cluster in an Amazon VPC, you need
            to create a s group before you start creating a cluster. For more
            information, see Subnets and Subnet Groups.

    CacheSecurityGroupNames
        A list of cache security group names to associate with this replication group.

    SecurityGroupIds
        One or more Amazon VPC security groups associated with this replication group.  Use this
        parameter only when you are creating a replication group in an VPC.

    Tags
        A list of tags to be added to this resource.  Note that due to shortcomings in the
        AWS API for Elasticache, these can only be set during resource creation - later
        modification is not (currently) supported.

    SnapshotArns
        A list of ARNs that uniquely identify the Redis RDB snapshot files stored in Amazon S3.
        These snapshot files are used to populate the replication group.  The Amazon S3 object name
        in the ARN cannot contain any commas. The list must match the number of node groups (shards)
        in the replication group, which means you cannot repartition.

        .. note::
            This parameter is only valid if the Engine parameter is redis.

    SnapshotName
        The name of a snapshot from which to restore data into the new replication group.  The
        snapshot status changes to restoring while the new replication group is being created.
        Note:  This parameter is only valid if the Engine parameter is redis.

    PreferredMaintenanceWindow
        Specifies the weekly time range during which maintenance on the cluster is performed. It is
        specified as a range in the format ddd:hh24:mi-ddd:hh24:mi (24H Clock UTC). The minimum
        maintenance window is a 60 minute period.
        Valid values for ddd are:  sun, mon, tue, wed, thu, fri, sat

        Example:  sun:23:00-mon:01:30

    Port
        The port number on which each member of the replication group accepts connections.

    NotificationTopicArn
        The ARN of an SNS topic to which notifications are sent.

        .. note::
            The SNS topic owner must be the same as the cache cluster owner.

    AutoMinorVersionUpgrade
        This parameter is currently disabled.

    SnapshotRetentionLimit
        The number of days for which ElastiCache will retain automatic snapshots before deleting
        them.

        Default:  0 (that is, automatic backups are disabled for this cache cluster).

        .. note::
            This parameter is only valid if the Engine parameter is redis.

    SnapshotWindow
        The daily time range (in UTC) during which ElastiCache begins taking a daily snapshot of
        your node group (shard).  If you do not specify this parameter, ElastiCache automatically
        chooses an appropriate time range.

        Example:  05:00-09:00

        .. note::
            This parameter is only valid if the Engine parameter is redis.

    AuthToken
        The password used to access a password protected server.
        Password constraints:

        - Must be only printable ASCII characters.
        - Must be at least 16 characters and no more than 128 characters in length.
        - Cannot contain any of the following characters: '/', '"', or "@".

    SnapshottingClusterId
        The cache cluster ID that is used as the daily snapshot source for the replication group.

    NotificationTopicStatus
        The status of the SNS notification topic.  Notifications are sent only if the status is active.
        Valid values:  active | inactive

    region
        Region to connect to.

    key
        Secret key to be used.

    keyid
        Access key to be used.

    profile
        A dict with region, key and keyid, or a pillar key (string)
        that contains a dict with region, key and keyid.
    '''
    ret = {'name': name, 'result': True, 'comment': '', 'changes': {}}
    args = dict([(k, v) for k, v in args.items() if not k.startswith('_')])
    current = __salt__['boto3_elasticache.'
                       'describe_replication_groups'](name, region=region, key=key,
                                                      keyid=keyid, profile=profile)
    if current:
        check_update = True
    else:
        check_update = False
        only_on_modify = [
            'SnapshottingClusterId',
            'NotificationTopicStatus'
        ]
        create_args = {}
        for k, v in args.items():
            if k in only_on_modify:
                check_update = True
            else:
                create_args[k] = v
        if __opts__['test']:
            ret['comment'] = 'Replication group {0} would be created.'.format(name)
            ret['result'] = None
            return ret
        created = __salt__['boto3_elasticache.'
                           'create_replication_group'](name, wait=wait,
                                                       security_groups=security_groups,
                                                       region=region, key=key, keyid=keyid,
                                                       profile=profile, **create_args)
        if created:
            new = __salt__['boto3_elasticache.'
                           'describe_replication_groups'](name, region=region, key=key,
                                                          keyid=keyid, profile=profile)
            ret['comment'] = 'Replication group {0} was created.'.format(name)
            ret['changes']['old'] = None
            ret['changes']['new'] = new[0]
        else:
            ret['result'] = False
            ret['comment'] = 'Failed to create {0} replication group.'.format(name)

    if check_update:
        # Refresh this in case we're updating from 'only_on_modify' above...
        updated = __salt__['boto3_elasticache.'
                           'describe_replication_groups'](name, region=region, key=key,
                                                          keyid=keyid, profile=profile)[0]
        need_update = _diff_replication_group(updated, args)
        if need_update:
            if __opts__['test']:
                ret['comment'] = 'Replication group {0} would be modified.'.format(name)
                ret['result'] = None
                return ret
            modified = __salt__['boto3_elasticache.'
                                'modify_replication_group'](name, wait=wait,
                                                            security_groups=security_groups,
                                                            region=region, key=key, keyid=keyid,
                                                            profile=profile, **need_update)
            if modified:
                new = __salt__['boto3_elasticache.'
                               'describe_replication_groups'](name, region=region, key=key,
                                                              keyid=keyid, profile=profile)
                if ret['comment']:  # 'create' just ran...
                    ret['comment'] += ' ... and then immediately modified.'
                else:
                    ret['comment'] = 'Replication group {0} was modified.'.format(name)
                    ret['changes']['old'] = current[0] if len(current) else None
                ret['changes']['new'] = new[0]
            else:
                ret['result'] = False
                ret['comment'] = 'Failed to modify replication group {0}.'.format(name)
        else:
            ret['comment'] = 'Replication group {0} is in the desired state.'.format(name)
    return ret


def replication_group_absent(name, wait=600, region=None, key=None, keyid=None,
                             profile=None, **args):
    '''
    Ensure a given replication group is deleted.

    name
        Name of the replication group.

    wait
        Integer describing how long, in seconds, to wait for confirmation from AWS that the
        resource is in the desired state.  Zero meaning to return success or failure immediately
        of course.  Note that waiting for the cluster to become available is generally the
        better course, as failure to do so will often lead to subsequent failures when managing
        dependent resources.

    ReplicationGroupId
        The replication group identifier.
        Note:  In general this parameter is not needed, as 'name' is used if it's not provided.

    RetainPrimaryCluster
        If set to true, all of the read replicas are deleted, but the primary node is retained.

    FinalSnapshotIdentifier
        The name of a final node group (shard) snapshot.  ElastiCache creates the snapshot from
        the primary node in the cluster, rather than one of the replicas; this is to ensure that
        it captures the freshest data.  After the final snapshot is taken, the replication group is
        immediately deleted.

    region
        Region to connect to.

    key
        Secret key to be used.

    keyid
        Access key to be used.

    profile
        A dict with region, key and keyid, or a pillar key (string)
        that contains a dict with region, key and keyid.
    '''
    ret = {'name': name, 'result': True, 'comment': '', 'changes': {}}
    args = dict([(k, v) for k, v in args.items() if not k.startswith('_')])
    exists = __salt__['boto3_elasticache.'
                      'replication_group_exists'](name, region=region, key=key,
                                                  keyid=keyid, profile=profile)
    if exists:
        if __opts__['test']:
            ret['comment'] = 'Replication group {0} would be removed.'.format(name)
            ret['result'] = None
            return ret
        deleted = __salt__['boto3_elasticache.'
                             'delete_replication_group'](name, wait=wait, region=region, key=key,
                                                       keyid=keyid, profile=profile, **args)
        if deleted:
            ret['changes']['old'] = name
            ret['changes']['new'] = None
        else:
            ret['result'] = False
            ret['comment'] = 'Failed to delete {0} replication group.'.format(name)
    else:
        ret['comment'] = 'Replication group {0} already absent.'.format(name)
    return ret


def _diff_cache_subnet_group(current, desired):
    '''
    If you need to enhance what modify_cache_subnet_group() considers when deciding what is to be
    (or can be) updated, add it to 'modifiable' below.  It's a dict mapping the param as used
    in modify_cache_subnet_group() to that in describe_cache_subnet_group().  Any data fiddlery that
    needs to be done to make the mappings meaningful should be done in the munging section
    below as well.

    This function will ONLY touch settings that are explicitly called out in 'desired' - any
    settings which might have previously been changed from their 'default' values will not be
    changed back simply by leaving them out of 'desired'.  This is both intentional, and
    much, much easier to code :)
    '''
    modifiable = {
        'CacheSubnetGroupDescription': 'CacheSubnetGroupDescription',
        'SubnetIds': 'SubnetIds'
    }

    need_update = {}
    for m, o in modifiable.items():
        if m in desired:
            if not o:
                # Always pass these through - let AWS do the math...
                need_update[m] = desired[m]
            else:
                if m in current:
                    # Equivalence testing works fine for current simple type comparisons
                    # This might need enhancement if more complex structures enter the picture
                    if current[m] != desired[m]:
                        need_update[m] = desired[m]
    return need_update


def cache_subnet_group_present(name, subnets=None, region=None, key=None, keyid=None, profile=None,
                               **args):
    '''
    Ensure cache subnet group exists.

    name
        A name for the cache subnet group. This value is stored as a lowercase string.
        Constraints:  Must contain no more than 255 alphanumeric characters or hyphens.

    subnets
        A list of VPC subnets (IDs, Names, or a mix) for the cache subnet group.

    CacheSubnetGroupName
        A name for the cache subnet group. This value is stored as a lowercase string.
        Constraints:  Must contain no more than 255 alphanumeric characters or hyphens.
        Note:  In general this parameter is not needed, as 'name' is used if it's not provided.

    CacheSubnetGroupDescription
        A description for the cache subnet group.

    SubnetIds
        A list of VPC subnet IDs for the cache subnet group.  This is ADDITIVE with 'subnets' above.

    region
        Region to connect to.

    key
        Secret key to be used.

    keyid
        Access key to be used.

    profile
        A dict with region, key and keyid, or a pillar key (string) that
        contains a dict with region, key and keyid.
    '''
    ret = {'name': name, 'result': True, 'comment': '', 'changes': {}}
    args = dict([(k, v) for k, v in args.items() if not k.startswith('_')])
    current = __salt__['boto3_elasticache.'
                       'describe_cache_subnet_groups'](name, region=region, key=key,
                                                       keyid=keyid, profile=profile)
    if current:
        check_update = True
    else:
        check_update = False
        if __opts__['test']:
            ret['comment'] = 'Cache subnet group {0} would be created.'.format(name)
            ret['result'] = None
            return ret
        created = __salt__['boto3_elasticache.'
                           'create_cache_subnet_group'](name, subnets=subnets,
                                                        region=region, key=key, keyid=keyid,
                                                        profile=profile, **args)
        if created:
            new = __salt__['boto3_elasticache.'
                           'describe_cache_subnet_groups'](name, region=region, key=key,
                                                           keyid=keyid, profile=profile)
            ret['comment'] = 'Cache subnet group {0} was created.'.format(name)
            ret['changes']['old'] = None
            ret['changes']['new'] = new[0]
        else:
            ret['result'] = False
            ret['comment'] = 'Failed to create {0} cache subnet group.'.format(name)

    if check_update:
        need_update = _diff_cache_subnet_group(current, args)
        if need_update:
            if __opts__['test']:
                ret['comment'] = 'Cache subnet group {0} would be modified.'.format(name)
                ret['result'] = None
                return ret
            modified = __salt__['boto3_elasticache.'
                                'modify_cache_subnet_group'](name, subnets=subnets,
                                                             region=region, key=key, keyid=keyid,
                                                             profile=profile, **need_update)
            if modified:
                new = __salt__['boto3_elasticache.'
                               'describe_cache_subnet_groups'](name, region=region, key=key,
                                                               keyid=keyid, profile=profile)
                ret['comment'] = 'Cache subnet group {0} was modified.'.format(name)
                ret['changes']['old'] = current['CacheSubetGroups'][0]
                ret['changes']['new'] = new[0]
            else:
                ret['result'] = False
                ret['comment'] = 'Failed to modify cache subnet group {0}.'.format(name)
        else:
            ret['comment'] = 'Cache subnet group {0} is in the desired state.'.format(name)
    return ret


def cache_subnet_group_absent(name, region=None, key=None, keyid=None, profile=None, **args):
    '''
    Ensure a given cache subnet group is deleted.

    name
        Name of the cache subnet group.

    CacheSubnetGroupName
        A name for the cache subnet group.
        Note:  In general this parameter is not needed, as 'name' is used if it's not provided.

    region
        Region to connect to.

    key
        Secret key to be used.

    keyid
        Access key to be used.

    profile
        A dict with region, key and keyid, or a pillar key (string)
        that contains a dict with region, key and keyid.
    '''
    ret = {'name': name, 'result': True, 'comment': '', 'changes': {}}
    args = dict([(k, v) for k, v in args.items() if not k.startswith('_')])
    exists = __salt__['boto3_elasticache.'
                      'cache_subnet_group_exists'](name, region=region, key=key,
                                                  keyid=keyid, profile=profile)
    if exists:
        if __opts__['test']:
            ret['comment'] = 'Cache subnet group {0} would be removed.'.format(name)
            ret['result'] = None
            return ret
        deleted = __salt__['boto3_elasticache.'
                           'delete_cache_subnet_group'](name, region=region, key=key,
                                                       keyid=keyid, profile=profile, **args)
        if deleted:
            ret['changes']['old'] = name
            ret['changes']['new'] = None
        else:
            ret['result'] = False
            ret['comment'] = 'Failed to delete {0} cache_subnet group.'.format(name)
    else:
        ret['comment'] = 'Cache subnet group {0} already absent.'.format(name)
    return ret



# -*- coding: utf-8 -*-
'''
Support for Portage

.. important::
    If you feel that Salt should be using this module to manage packages on a
    minion, and it is using a different module (or gives an error similar to
    *'pkg.install' is not available*), see :ref:`here
    <module-provider-override>`.

:optdepends:    - portage Python adapter

For now all package names *MUST* include the package category,
i.e. ``'vim'`` will not work, ``'app-editors/vim'`` will.
'''
from __future__ import absolute_import, print_function, unicode_literals

# Import python libs
import copy
import logging
import re

# Import salt libs
import salt.utils.args
import salt.utils.data
import salt.utils.functools
import salt.utils.path
import salt.utils.pkg
import salt.utils.systemd
import salt.utils.versions
from salt.exceptions import CommandExecutionError, MinionError
from salt.ext import six

# Import third party libs
HAS_PORTAGE = False
try:
    import portage
    HAS_PORTAGE = True
except ImportError:
    import os
    import sys
    if os.path.isdir('/usr/lib/portage/pym'):
        try:
            # In a virtualenv, the portage python path needs to be manually added
            sys.path.insert(0, '/usr/lib/portage/pym')
            import portage
            HAS_PORTAGE = True
        except ImportError:
            pass

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = 'pkg'


def __virtual__():
    '''
    Confirm this module is on a Gentoo based system
    '''
    if HAS_PORTAGE and __grains__['os'] == 'Gentoo':
        return __virtualname__
    return (False, 'The ebuild execution module cannot be loaded: either the system is not Gentoo or the portage python library is not available.')


def _vartree():
    import portage  # pylint: disable=3rd-party-module-not-gated
    portage = reload(portage)
    return portage.db[portage.root]['vartree']


def _porttree():
    import portage  # pylint: disable=3rd-party-module-not-gated
    portage = reload(portage)
    return portage.db[portage.root]['porttree']


def _p_to_cp(p):
    try:
        ret = portage.dep_getkey(p)
        if ret:
            return ret
    except portage.exception.InvalidAtom:
        pass

    try:
        ret = _porttree().dbapi.xmatch('bestmatch-visible', p)
        if ret:
            return portage.dep_getkey(ret)
    except portage.exception.InvalidAtom:
        pass

    try:
        ret = _porttree().dbapi.xmatch("match-all", p)
        if ret:
            return portage.cpv_getkey(ret[0])
    except portage.exception.InvalidAtom:
        pass

    return None


def _allnodes():
    if 'portage._allnodes' in __context__:
        return __context__['portage._allnodes']
    else:
        ret = _porttree().getallnodes()
        __context__['portage._allnodes'] = ret
        return ret


def _cpv_to_cp(cpv):
    try:
        ret = portage.dep_getkey(cpv)
        if ret:
            return ret
    except portage.exception.InvalidAtom:
        pass

    try:
        ret = portage.cpv_getkey(cpv)
        if ret:
            return ret
    except portage.exception.InvalidAtom:
        pass

    return cpv


def _cpv_to_version(cpv):
    return portage.versions.cpv_getversion(cpv)


def _process_emerge_err(stdout, stderr):
    '''
    Used to parse emerge output to provide meaningful output when emerge fails
    '''
    ret = {}
    rexp = re.compile(r'^[<>=][^ ]+/[^ ]+ [^\n]+', re.M)

    slot_conflicts = re.compile(r'^[^ \n]+/[^ ]+:[^ ]', re.M).findall(stderr)
    if slot_conflicts:
        ret['slot conflicts'] = slot_conflicts

    blocked = re.compile(r'(?m)^\[blocks .+\] '
                         r'([^ ]+/[^ ]+-[0-9]+[^ ]+)'
                         r'.*$').findall(stdout)

    unsatisfied = re.compile(
            r'Error: The above package list contains').findall(stderr)

    # If there were blocks and emerge could not resolve it.
    if blocked and unsatisfied:
        ret['blocked'] = blocked

    sections = re.split('\n\n', stderr)
    for section in sections:
        if 'The following keyword changes' in section:
            ret['keywords'] = rexp.findall(section)
        elif 'The following license changes' in section:
            ret['license'] = rexp.findall(section)
        elif 'The following USE changes' in section:
            ret['use'] = rexp.findall(section)
        elif 'The following mask changes' in section:
            ret['mask'] = rexp.findall(section)
    return ret


def check_db(*names, **kwargs):
    '''
    .. versionadded:: 0.17.0

    Returns a dict containing the following information for each specified
    package:

    1. A key ``found``, which will be a boolean value denoting if a match was
       found in the package database.
    2. If ``found`` is ``False``, then a second key called ``suggestions`` will
       be present, which will contain a list of possible matches. This list
       will be empty if the package name was specified in ``category/pkgname``
       format, since the suggestions are only intended to disambiguate
       ambiguous package names (ones submitted without a category).

    CLI Examples:

    .. code-block:: bash

        salt '*' pkg.check_db <package1> <package2> <package3>
    '''
    ### NOTE: kwargs is not used here but needs to be present due to it being
    ### required in the check_db function in other package providers.
    ret = {}
    for name in names:
        if name in ret:
            log.warning('pkg.check_db: Duplicate package name \'{0}\' '
                        'submitted'.format(name))
            continue
        if '/' not in name:
            ret.setdefault(name, {})['found'] = False
            ret[name]['suggestions'] = porttree_matches(name)
        else:
            ret.setdefault(name, {})['found'] = name in _allnodes()
            if ret[name]['found'] is False:
                ret[name]['suggestions'] = []
    return ret


def ex_mod_init(low):
    '''
    If the config option ``ebuild.enforce_nice_config`` is set to True, this
    module will enforce a nice tree structure for /etc/portage/package.*
    configuration files.

    .. versionadded:: 0.17.0
       Initial automatic enforcement added when pkg is used on a Gentoo system.

    .. versionchanged:: 2014.1.0-Hydrogen
       Configure option added to make this behaviour optional, defaulting to
       off.

    .. seealso::
       ``ebuild.ex_mod_init`` is called automatically when a state invokes a
       pkg state on a Gentoo system.
       :py:func:`salt.states.pkg.mod_init`

       ``ebuild.ex_mod_init`` uses ``portage_config.enforce_nice_config`` to do
       the lifting.
       :py:func:`salt.modules.portage_config.enforce_nice_config`

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.ex_mod_init
    '''
    if __salt__['config.get']('ebuild.enforce_nice_config', False):
        __salt__['portage_config.enforce_nice_config']()
    return True


def latest_version(*names, **kwargs):
    '''
    Return the latest version of the named package available for upgrade or
    installation. If more than one package name is specified, a dict of
    name/version pairs is returned.

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.latest_version <package name>
        salt '*' pkg.latest_version <package1> <package2> <package3> ...
    '''
    refresh = salt.utils.data.is_true(kwargs.pop('refresh', True))

    if len(names) == 0:
        return ''

    # Refresh before looking for the latest version available
    if refresh:
        refresh_db()

    ret = {}
    # Initialize the dict with empty strings
    for name in names:
        ret[name] = ''
        installed = _cpv_to_version(_vartree().dep_bestmatch(name))
        avail = _cpv_to_version(_porttree().dep_bestmatch(name))
        if avail and (not installed or salt.utils.versions.compare(ver1=installed, oper='<', ver2=avail, cmp_func=version_cmp)):
            ret[name] = avail

    # Return a string if only one package name passed
    if len(names) == 1:
        return ret[names[0]]
    return ret

# available_version is being deprecated
available_version = salt.utils.functools.alias_function(latest_version, 'available_version')


def _get_upgradable(backtrack=3):
    '''
    Utility function to get upgradable packages

    Sample return data:
    { 'pkgname': '1.2.3-45', ... }
    '''

    cmd = ['emerge',
           '--ask', 'n',
           '--backtrack', '{0}'.format(backtrack),
           '--pretend',
           '--update',
           '--newuse',
           '--deep',
           '@world']

    call = __salt__['cmd.run_all'](cmd,
                                   output_loglevel='trace',
                                   python_shell=False)

    if call['retcode'] != 0:
        msg = 'Failed to get upgrades'
        for key in ('stderr', 'stdout'):
            if call[key]:
                msg += ': ' + call[key]
                break
        raise CommandExecutionError(msg)
    else:
        out = call['stdout']

    rexp = re.compile(r'(?m)^\[.+\] '
                      r'([^ ]+/[^ ]+)'    # Package string
                      '-'
                      r'([0-9]+[^ ]+)'          # Version
                      r'.*$')
    keys = ['name', 'version']
    _get = lambda l, k: l[keys.index(k)]

    upgrades = rexp.findall(out)

    ret = {}
    for line in upgrades:
        name = _get(line, 'name')
        version_num = _get(line, 'version')
        ret[name] = version_num

    return ret


def list_upgrades(refresh=True, backtrack=3, **kwargs):  # pylint: disable=W0613
    '''
    List all available package upgrades.

    refresh
        Whether or not to sync the portage tree before checking for upgrades.

    backtrack
        Specifies an integer number of times to backtrack if dependency
        calculation fails due to a conflict or an unsatisfied dependency
        (default: ´3´).

        .. versionadded: 2015.8.0

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.list_upgrades
    '''
    if salt.utils.data.is_true(refresh):
        refresh_db()
    return _get_upgradable(backtrack)


def upgrade_available(name):
    '''
    Check whether or not an upgrade is available for a given package

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.upgrade_available <package name>
    '''
    return latest_version(name) != ''


def version(*names, **kwargs):
    '''
    Returns a string representing the package version or an empty string if not
    installed. If more than one package name is specified, a dict of
    name/version pairs is returned.

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.version <package name>
        salt '*' pkg.version <package1> <package2> <package3> ...
    '''
    return __salt__['pkg_resource.version'](*names, **kwargs)


def porttree_matches(name):
    '''
    Returns a list containing the matches for a given package name from the
    portage tree. Note that the specific version of the package will not be
    provided for packages that have several versions in the portage tree, but
    rather the name of the package (i.e. "dev-python/paramiko").
    '''
    matches = []
    for category in _porttree().dbapi.categories:
        if _porttree().dbapi.cp_list(category + "/" + name):
            matches.append(category + "/" + name)
    return matches


def list_pkgs(versions_as_list=False, **kwargs):
    '''
    List the packages currently installed in a dict::

        {'<package_name>': '<version>'}

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.list_pkgs
    '''
    versions_as_list = salt.utils.data.is_true(versions_as_list)
    # not yet implemented or not applicable
    if any([salt.utils.data.is_true(kwargs.get(x))
            for x in ('removed', 'purge_desired')]):
        return {}

    if 'pkg.list_pkgs' in __context__:
        if versions_as_list:
            return __context__['pkg.list_pkgs']
        else:
            ret = copy.deepcopy(__context__['pkg.list_pkgs'])
            __salt__['pkg_resource.stringify'](ret)
            return ret

    ret = {}
    pkgs = _vartree().dbapi.cpv_all()
    for cpv in pkgs:
        __salt__['pkg_resource.add_pkg'](ret,
                                         _cpv_to_cp(cpv),
                                         _cpv_to_version(cpv))
    __salt__['pkg_resource.sort_pkglist'](ret)
    __context__['pkg.list_pkgs'] = copy.deepcopy(ret)
    if not versions_as_list:
        __salt__['pkg_resource.stringify'](ret)
    return ret


def refresh_db():
    '''
    Updates the portage tree (emerge --sync). Uses eix-sync if available.

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.refresh_db
    '''
    # Remove rtag file to keep multiple refreshes from happening in pkg states
    salt.utils.pkg.clear_rtag(__opts__)
    if 'eix.sync' in __salt__:
        return __salt__['eix.sync']()

    if 'makeconf.features_contains'in __salt__ and __salt__['makeconf.features_contains']('webrsync-gpg'):
        # GPG sign verify is supported only for "webrsync"
        cmd = 'emerge-webrsync -q'
        # We prefer 'delta-webrsync' to 'webrsync'
        if salt.utils.path.which('emerge-delta-webrsync'):
            cmd = 'emerge-delta-webrsync -q'
        return __salt__['cmd.retcode'](cmd, python_shell=False) == 0
    else:
        if __salt__['cmd.retcode']('emerge --ask n --quiet --sync',
                                   python_shell=False) == 0:
            return True
        # We fall back to "webrsync" if "rsync" fails for some reason
        cmd = 'emerge-webrsync -q'
        # We prefer 'delta-webrsync' to 'webrsync'
        if salt.utils.path.which('emerge-delta-webrsync'):
            cmd = 'emerge-delta-webrsync -q'
        return __salt__['cmd.retcode'](cmd, python_shell=False) == 0


def _flags_changed(inst_flags, conf_flags):
    '''
    @type inst_flags: list
    @param inst_flags: list of use flags which were used
        when package was installed
    @type conf_flags: list
    @param conf_flags: list of use flags form portage/package.use
    @rtype: bool
    @return: True, if lists have changes
    '''
    conf_flags = conf_flags[:]
    for i in inst_flags:
        try:
            conf_flags.remove(i)
        except ValueError:
            return True
    return True if conf_flags else False


def install(name=None,
            refresh=False,
            pkgs=None,
            sources=None,
            slot=None,
            fromrepo=None,
            uses=None,
            binhost=None,
            **kwargs):
    '''
    .. versionchanged:: 2015.8.12,2016.3.3,2016.11.0
        On minions running systemd>=205, `systemd-run(1)`_ is now used to
        isolate commands which modify installed packages from the
        ``salt-minion`` daemon's control group. This is done to keep systemd
        from killing any emerge commands spawned by Salt when the
        ``salt-minion`` service is restarted. (see ``KillMode`` in the
        `systemd.kill(5)`_ manpage for more information). If desired, usage of
        `systemd-run(1)`_ can be suppressed by setting a :mod:`config option
        <salt.modules.config.get>` called ``systemd.scope``, with a value of
        ``False`` (no quotes).

    .. _`systemd-run(1)`: https://www.freedesktop.org/software/systemd/man/systemd-run.html
    .. _`systemd.kill(5)`: https://www.freedesktop.org/software/systemd/man/systemd.kill.html

    Install the passed package(s), add refresh=True to sync the portage tree
    before package is installed.

    name
        The name of the package to be installed. Note that this parameter is
        ignored if either "pkgs" or "sources" is passed. Additionally, please
        note that this option can only be used to emerge a package from the
        portage tree. To install a tbz2 package manually, use the "sources"
        option described below.

        CLI Example:

        .. code-block:: bash

            salt '*' pkg.install <package name>

    refresh
        Whether or not to sync the portage tree before installing.

    version
        Install a specific version of the package, e.g. 1.0.9-r1. Ignored
        if "pkgs" or "sources" is passed.

    slot
        Similar to version, but specifies a valid slot to be installed. It
        will install the latest available version in the specified slot.
        Ignored if "pkgs" or "sources" or "version" is passed.

        CLI Example:

        .. code-block:: bash

            salt '*' pkg.install sys-devel/gcc slot='4.4'

    fromrepo
        Similar to slot, but specifies the repository from the package will be
        installed. It will install the latest available version in the
        specified repository.
        Ignored if "pkgs" or "sources" or "version" is passed.

        CLI Example:

        .. code-block:: bash

            salt '*' pkg.install salt fromrepo='gentoo'

    uses
        Similar to slot, but specifies a list of use flag.
        Ignored if "pkgs" or "sources" or "version" is passed.

        CLI Example:

        .. code-block:: bash

            salt '*' pkg.install sys-devel/gcc uses='["nptl","-nossp"]'


    Multiple Package Installation Options:

    pkgs
        A list of packages to install from the portage tree. Must be passed as
        a python list.

        CLI Example:

        .. code-block:: bash

            salt '*' pkg.install pkgs='["foo","bar","~category/package:slot::repository[use]"]'

    sources
        A list of tbz2 packages to install. Must be passed as a list of dicts,
        with the keys being package names, and the values being the source URI
        or local path to the package.

        CLI Example:

        .. code-block:: bash

            salt '*' pkg.install sources='[{"foo": "salt://foo.tbz2"},{"bar": "salt://bar.tbz2"}]'
    binhost
        has two options try and force.
        try - tells emerge to try and install the package from a configured binhost.
        force - forces emerge to install the package from a binhost otherwise it fails out.

    Returns a dict containing the new package names and versions::

        {'<package>': {'old': '<old-version>',
                       'new': '<new-version>'}}
    '''
    log.debug('Called modules.pkg.install: {0}'.format(
        {
            'name': name,
            'refresh': refresh,
            'pkgs': pkgs,
            'sources': sources,
            'kwargs': kwargs,
            'binhost': binhost,
        }
    ))
    if salt.utils.data.is_true(refresh):
        refresh_db()

    try:
        pkg_params, pkg_type = __salt__['pkg_resource.parse_targets'](
            name, pkgs, sources, **kwargs
        )
    except MinionError as exc:
        raise CommandExecutionError(exc)

    # Handle version kwarg for a single package target
    if pkgs is None and sources is None:
        version_num = kwargs.get('version')
        if not version_num:
            version_num = ''
            if slot is not None:
                version_num += ':{0}'.format(slot)
            if fromrepo is not None:
                version_num += '::{0}'.format(fromrepo)
            if uses is not None:
                version_num += '[{0}]'.format(','.join(uses))
            pkg_params = {name: version_num}

    if pkg_params is None or len(pkg_params) == 0:
        return {}
    elif pkg_type == 'file':
        emerge_opts = ['tbz2file']
    else:
        emerge_opts = []

    if binhost == 'try':
        bin_opts = ['-g']
    elif binhost == 'force':
        bin_opts = ['-G']
    else:
        bin_opts = []

    changes = {}

    if pkg_type == 'repository':
        targets = list()
        for param, version_num in six.iteritems(pkg_params):
            original_param = param
            param = _p_to_cp(param)
            if param is None:
                raise portage.dep.InvalidAtom(original_param)

            if version_num is None:
                targets.append(param)
            else:
                keyword = None

                match = re.match('^(~)?([<>])?(=)?([^<>=]*)$', version_num)
                if match:
                    keyword, gt_lt, eq, verstr = match.groups()
                    prefix = gt_lt or ''
                    prefix += eq or ''
                    # If no prefix characters were supplied and verstr contains a version, use '='
                    if len(verstr) > 0 and verstr[0] != ':' and verstr[0] != '[':
                        prefix = prefix or '='
                        target = '{0}{1}-{2}'.format(prefix, param, verstr)
                    else:
                        target = '{0}{1}'.format(param, verstr)
                else:
                    target = '{0}'.format(param)

                if '[' in target:
                    old = __salt__['portage_config.get_flags_from_package_conf']('use', target)
                    __salt__['portage_config.append_use_flags'](target)
                    new = __salt__['portage_config.get_flags_from_package_conf']('use', target)
                    if old != new:
                        changes[param + '-USE'] = {'old': old, 'new': new}
                    target = target[:target.rfind('[')]

                if keyword is not None:
                    __salt__['portage_config.append_to_package_conf']('accept_keywords',
                                                                        target,
                                                                        ['~ARCH'])
                    changes[param + '-ACCEPT_KEYWORD'] = {'old': '', 'new': '~ARCH'}

                if not changes:
                    inst_v = version(param)

                    # Prevent latest_version from calling refresh_db. Either we
                    # just called it or we were asked not to.
                    if latest_version(param, refresh=False) == inst_v:
                        all_uses = __salt__['portage_config.get_cleared_flags'](param)
                        if _flags_changed(*all_uses):
                            changes[param] = {'version': inst_v,
                                              'old': {'use': all_uses[0]},
                                              'new': {'use': all_uses[1]}}
                targets.append(target)
    else:
        targets = pkg_params

    cmd = []
    if salt.utils.systemd.has_scope(__context__) \
            and __salt__['config.get']('systemd.scope', True):
        cmd.extend(['systemd-run', '--scope'])
    cmd.extend(['emerge', '--ask', 'n', '--quiet'])
    cmd.extend(bin_opts)
    cmd.extend(emerge_opts)
    cmd.extend(targets)

    old = list_pkgs()
    call = __salt__['cmd.run_all'](cmd,
                                   output_loglevel='trace',
                                   python_shell=False)
    if call['retcode'] != 0:
        needed_changes = _process_emerge_err(call['stdout'], call['stderr'])
    else:
        needed_changes = []

    __context__.pop('pkg.list_pkgs', None)
    new = list_pkgs()
    changes.update(salt.utils.data.compare_dicts(old, new))

    if needed_changes:
        raise CommandExecutionError(
            'Error occurred installing package(s)',
            info={'needed changes': needed_changes, 'changes': changes}
        )

    return changes


def update(pkg, slot=None, fromrepo=None, refresh=False, binhost=None):
    '''
    .. versionchanged:: 2015.8.12,2016.3.3,2016.11.0
        On minions running systemd>=205, `systemd-run(1)`_ is now used to
        isolate commands which modify installed packages from the
        ``salt-minion`` daemon's control group. This is done to keep systemd
        from killing any emerge commands spawned by Salt when the
        ``salt-minion`` service is restarted. (see ``KillMode`` in the
        `systemd.kill(5)`_ manpage for more information). If desired, usage of
        `systemd-run(1)`_ can be suppressed by setting a :mod:`config option
        <salt.modules.config.get>` called ``systemd.scope``, with a value of
        ``False`` (no quotes).

    .. _`systemd-run(1)`: https://www.freedesktop.org/software/systemd/man/systemd-run.html
    .. _`systemd.kill(5)`: https://www.freedesktop.org/software/systemd/man/systemd.kill.html

    Updates the passed package (emerge --update package)

    slot
        Restrict the update to a particular slot. It will update to the
        latest version within the slot.

    fromrepo
        Restrict the update to a particular repository. It will update to the
        latest version within the repository.
    binhost
        has two options try and force.
        try - tells emerge to try and install the package from a configured binhost.
        force - forces emerge to install the package from a binhost otherwise it fails out.

    Return a dict containing the new package names and versions::

        {'<package>': {'old': '<old-version>',
                       'new': '<new-version>'}}

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.update <package name>
    '''
    if salt.utils.data.is_true(refresh):
        refresh_db()

    full_atom = pkg

    if slot is not None:
        full_atom = '{0}:{1}'.format(full_atom, slot)

    if fromrepo is not None:
        full_atom = '{0}::{1}'.format(full_atom, fromrepo)

    if binhost == 'try':
        bin_opts = ['-g']
    elif binhost == 'force':
        bin_opts = ['-G']
    else:
        bin_opts = []

    old = list_pkgs()
    cmd = []
    if salt.utils.systemd.has_scope(__context__) \
            and __salt__['config.get']('systemd.scope', True):
        cmd.extend(['systemd-run', '--scope'])
    cmd.extend(['emerge',
                '--ask', 'n',
                '--quiet',
                '--update',
                '--newuse',
                '--oneshot'])
    cmd.extend(bin_opts)
    cmd.append(full_atom)
    call = __salt__['cmd.run_all'](cmd,
                                   output_loglevel='trace',
                                   python_shell=False)
    if call['retcode'] != 0:
        needed_changes = _process_emerge_err(call['stdout'], call['stderr'])
    else:
        needed_changes = []

    __context__.pop('pkg.list_pkgs', None)
    new = list_pkgs()
    ret = salt.utils.data.compare_dicts(old, new)

    if needed_changes:
        raise CommandExecutionError(
            'Problem encountered updating package(s)',
            info={'needed_changes': needed_changes, 'changes': ret}
        )

    return ret


def upgrade(refresh=True, binhost=None, backtrack=3):
    '''
    .. versionchanged:: 2015.8.12,2016.3.3,2016.11.0
        On minions running systemd>=205, `systemd-run(1)`_ is now used to
        isolate commands which modify installed packages from the
        ``salt-minion`` daemon's control group. This is done to keep systemd
        from killing any emerge commands spawned by Salt when the
        ``salt-minion`` service is restarted. (see ``KillMode`` in the
        `systemd.kill(5)`_ manpage for more information). If desired, usage of
        `systemd-run(1)`_ can be suppressed by setting a :mod:`config option
        <salt.modules.config.get>` called ``systemd.scope``, with a value of
        ``False`` (no quotes).

    .. _`systemd-run(1)`: https://www.freedesktop.org/software/systemd/man/systemd-run.html
    .. _`systemd.kill(5)`: https://www.freedesktop.org/software/systemd/man/systemd.kill.html

    Run a full system upgrade (emerge -uDN @world)

    binhost
        has two options try and force.
        try - tells emerge to try and install the package from a configured binhost.
        force - forces emerge to install the package from a binhost otherwise it fails out.

    backtrack
        Specifies an integer number of times to backtrack if dependency
        calculation fails due to a conflict or an unsatisfied dependency
        (default: ´3´).

        .. versionadded: 2015.8.0

    Returns a dictionary containing the changes:

    .. code-block:: python

        {'<package>':  {'old': '<old-version>',
                        'new': '<new-version>'}}


    CLI Example:

    .. code-block:: bash

        salt '*' pkg.upgrade
    '''
    ret = {'changes': {},
           'result': True,
           'comment': ''}

    if salt.utils.data.is_true(refresh):
        refresh_db()

    if binhost == 'try':
        bin_opts = ['--getbinpkg']
    elif binhost == 'force':
        bin_opts = ['--getbinpkgonly']
    else:
        bin_opts = []

    old = list_pkgs()
    cmd = []
    if salt.utils.systemd.has_scope(__context__) \
            and __salt__['config.get']('systemd.scope', True):
        cmd.extend(['systemd-run', '--scope'])
    cmd.extend(['emerge',
                '--ask', 'n',
                '--quiet',
                '--backtrack', '{0}'.format(backtrack),
                '--update',
                '--newuse',
                '--deep'])
    if bin_opts:
        cmd.extend(bin_opts)
    cmd.append('@world')

    result = __salt__['cmd.run_all'](cmd,
                                     output_loglevel='trace',
                                     python_shell=False)
    __context__.pop('pkg.list_pkgs', None)
    new = list_pkgs()
    ret = salt.utils.data.compare_dicts(old, new)

    if result['retcode'] != 0:
        raise CommandExecutionError(
            'Problem encountered upgrading packages',
            info={'changes': ret, 'result': result}
        )

    return ret


def remove(name=None, slot=None, fromrepo=None, pkgs=None, **kwargs):
    '''
    .. versionchanged:: 2015.8.12,2016.3.3,2016.11.0
        On minions running systemd>=205, `systemd-run(1)`_ is now used to
        isolate commands which modify installed packages from the
        ``salt-minion`` daemon's control group. This is done to keep systemd
        from killing any emerge commands spawned by Salt when the
        ``salt-minion`` service is restarted. (see ``KillMode`` in the
        `systemd.kill(5)`_ manpage for more information). If desired, usage of
        `systemd-run(1)`_ can be suppressed by setting a :mod:`config option
        <salt.modules.config.get>` called ``systemd.scope``, with a value of
        ``False`` (no quotes).

    .. _`systemd-run(1)`: https://www.freedesktop.org/software/systemd/man/systemd-run.html
    .. _`systemd.kill(5)`: https://www.freedesktop.org/software/systemd/man/systemd.kill.html

    Remove packages via emerge --unmerge.

    name
        The name of the package to be deleted.

    slot
        Restrict the remove to a specific slot. Ignored if ``name`` is None.

    fromrepo
        Restrict the remove to a specific slot. Ignored if ``name`` is None.

    Multiple Package Options:

    pkgs
        Uninstall multiple packages. ``slot`` and ``fromrepo`` arguments are
        ignored if this argument is present. Must be passed as a python list.

    .. versionadded:: 0.16.0

    Returns a dict containing the changes.

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.remove <package name>
        salt '*' pkg.remove <package name> slot=4.4 fromrepo=gentoo
        salt '*' pkg.remove <package1>,<package2>,<package3>
        salt '*' pkg.remove pkgs='["foo", "bar"]'
    '''
    try:
        pkg_params = __salt__['pkg_resource.parse_targets'](name, pkgs)[0]
    except MinionError as exc:
        raise CommandExecutionError(exc)

    old = list_pkgs()
    if name and not pkgs and (slot is not None or fromrepo is not None)and len(pkg_params) == 1:
        fullatom = name
        if slot is not None:
            targets = ['{0}:{1}'.format(fullatom, slot)]
        if fromrepo is not None:
            targets = ['{0}::{1}'.format(fullatom, fromrepo)]
        targets = [fullatom]
    else:
        targets = [x for x in pkg_params if x in old]

    if not targets:
        return {}

    cmd = []
    if salt.utils.systemd.has_scope(__context__) \
            and __salt__['config.get']('systemd.scope', True):
        cmd.extend(['systemd-run', '--scope'])
    cmd.extend(['emerge',
                '--ask', 'n',
                '--quiet',
                '--unmerge',
                '--quiet-unmerge-warn'])
    cmd.extend(targets)

    out = __salt__['cmd.run_all'](
        cmd,
        output_loglevel='trace',
        python_shell=False
    )
    if out['retcode'] != 0 and out['stderr']:
        errors = [out['stderr']]
    else:
        errors = []

    __context__.pop('pkg.list_pkgs', None)
    new = list_pkgs()
    ret = salt.utils.data.compare_dicts(old, new)

    if errors:
        raise CommandExecutionError(
            'Problem encountered removing package(s)',
            info={'errors': errors, 'changes': ret}
        )

    return ret


def purge(name=None, slot=None, fromrepo=None, pkgs=None, **kwargs):
    '''
    .. versionchanged:: 2015.8.12,2016.3.3,2016.11.0
        On minions running systemd>=205, `systemd-run(1)`_ is now used to
        isolate commands which modify installed packages from the
        ``salt-minion`` daemon's control group. This is done to keep systemd
        from killing any emerge commands spawned by Salt when the
        ``salt-minion`` service is restarted. (see ``KillMode`` in the
        `systemd.kill(5)`_ manpage for more information). If desired, usage of
        `systemd-run(1)`_ can be suppressed by setting a :mod:`config option
        <salt.modules.config.get>` called ``systemd.scope``, with a value of
        ``False`` (no quotes).

    .. _`systemd-run(1)`: https://www.freedesktop.org/software/systemd/man/systemd-run.html
    .. _`systemd.kill(5)`: https://www.freedesktop.org/software/systemd/man/systemd.kill.html

    Portage does not have a purge, this function calls remove followed
    by depclean to emulate a purge process

    name
        The name of the package to be deleted.

    slot
        Restrict the remove to a specific slot. Ignored if name is None.

    fromrepo
        Restrict the remove to a specific slot. Ignored if ``name`` is None.

    Multiple Package Options:

    pkgs
        Uninstall multiple packages. ``slot`` and ``fromrepo`` arguments are
        ignored if this argument is present. Must be passed as a python list.

    .. versionadded:: 0.16.0


    Returns a dict containing the changes.

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.purge <package name>
        salt '*' pkg.purge <package name> slot=4.4
        salt '*' pkg.purge <package1>,<package2>,<package3>
        salt '*' pkg.purge pkgs='["foo", "bar"]'
    '''
    ret = remove(name=name, slot=slot, fromrepo=fromrepo, pkgs=pkgs)
    ret.update(depclean(name=name, slot=slot, fromrepo=fromrepo, pkgs=pkgs))
    return ret


def depclean(name=None, slot=None, fromrepo=None, pkgs=None):
    '''
    Portage has a function to remove unused dependencies. If a package
    is provided, it will only removed the package if no other package
    depends on it.

    name
        The name of the package to be cleaned.

    slot
        Restrict the remove to a specific slot. Ignored if ``name`` is None.

    fromrepo
        Restrict the remove to a specific slot. Ignored if ``name`` is None.

    pkgs
        Clean multiple packages. ``slot`` and ``fromrepo`` arguments are
        ignored if this argument is present. Must be passed as a python list.

    Return a list containing the removed packages:

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.depclean <package name>
    '''
    try:
        pkg_params = __salt__['pkg_resource.parse_targets'](name, pkgs)[0]
    except MinionError as exc:
        raise CommandExecutionError(exc)

    old = list_pkgs()
    if name and not pkgs and (slot is not None or fromrepo is not None)and len(pkg_params) == 1:
        fullatom = name
        if slot is not None:
            targets = ['{0}:{1}'.format(fullatom, slot)]
        if fromrepo is not None:
            targets = ['{0}::{1}'.format(fullatom, fromrepo)]
        targets = [fullatom]
    else:
        targets = [x for x in pkg_params if x in old]

    cmd = ['emerge', '--ask', 'n', '--quiet', '--depclean'] + targets
    __salt__['cmd.run_all'](cmd,
                            output_loglevel='trace',
                            python_shell=False)
    __context__.pop('pkg.list_pkgs', None)
    new = list_pkgs()
    return salt.utils.data.compare_dicts(old, new)


def version_cmp(pkg1, pkg2, **kwargs):
    '''
    Do a cmp-style comparison on two packages. Return -1 if pkg1 < pkg2, 0 if
    pkg1 == pkg2, and 1 if pkg1 > pkg2. Return None if there was a problem
    making the comparison.

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.version_cmp '0.2.4-0' '0.2.4.1-0'
    '''
    # ignore_epoch is not supported here, but has to be included for API
    # compatibility. Rather than putting this argument into the function
    # definition (and thus have it show up in the docs), we just pop it out of
    # the kwargs dict and then raise an exception if any kwargs other than
    # ignore_epoch were passed.
    kwargs = salt.utils.args.clean_kwargs(**kwargs)
    kwargs.pop('ignore_epoch', None)
    if kwargs:
        salt.utils.args.invalid_kwargs(kwargs)

    regex = r'^~?([^:\[]+):?[^\[]*\[?.*$'
    ver1 = re.match(regex, pkg1)
    ver2 = re.match(regex, pkg2)

    if ver1 and ver2:
        return portage.versions.vercmp(ver1.group(1), ver2.group(1))
    return None


def version_clean(version):
    '''
    Clean the version string removing extra data.

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.version_clean <version_string>
    '''
    return re.match(r'^~?[<>]?=?([^<>=:\[]+).*$', version)


def check_extra_requirements(pkgname, pkgver):
    '''
    Check if the installed package already has the given requirements.

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.check_extra_requirements 'sys-devel/gcc' '~>4.1.2:4.1::gentoo[nls,fortran]'
    '''
    keyword = None

    match = re.match('^(~)?([<>])?(=)?([^<>=]*)$', pkgver)
    if match:
        keyword, gt_lt, eq, verstr = match.groups()
        prefix = gt_lt or ''
        prefix += eq or ''
        # We need to delete quotes around use flag list elements
        verstr = verstr.replace("'", "")
        # If no prefix characters were supplied and verstr contains a version, use '='
        if verstr[0] != ':' and verstr[0] != '[':
            prefix = prefix or '='
            atom = '{0}{1}-{2}'.format(prefix, pkgname, verstr)
        else:
            atom = '{0}{1}'.format(pkgname, verstr)
    else:
        return True

    try:
        cpv = _porttree().dbapi.xmatch('bestmatch-visible', atom)
    except portage.exception.InvalidAtom as iae:
        log.error('Unable to find a matching package for {0}: ({1})'.format(atom, iae))
        return False

    if cpv == '':
        return False

    try:
        cur_repo, cur_use = _vartree().dbapi.aux_get(cpv, ['repository', 'USE'])
    except KeyError:
        return False

    des_repo = re.match(r'^.+::([^\[]+).*$', atom)
    if des_repo and des_repo.group(1) != cur_repo:
        return False

    des_uses = set(portage.dep.dep_getusedeps(atom))
    cur_use = cur_use.split()
    if len([x for x in des_uses.difference(cur_use)
            if x[0] != '-' or x[1:] in cur_use]) > 0:
        return False

    if keyword:
        if not __salt__['portage_config.has_flag']('accept_keywords', atom, '~ARCH'):
            return False

    return True



# -*- coding: utf-8 -*-
'''
Sending Messages via SMTP
==========================

.. versionadded:: 2014.7.0

This state is useful for firing messages during state runs, using the SMTP
protocol

.. code-block:: yaml

    server-warning-message:
      smtp.send_msg:
        - name: 'This is a server warning message'
        - profile: my-smtp-account
        - recipient: admins@example.com
'''
# Import Python libs
from __future__ import absolute_import, unicode_literals, print_function


def __virtual__():
    '''
    Only load if the SMTP module is available in __salt__
    '''
    return 'smtp' if 'smtp.send_msg' in __salt__ else False


def send_msg(name, recipient, subject, sender=None, profile=None, use_ssl='True'):
    '''
    Send a message via SMTP

    .. code-block:: yaml

        server-warning-message:
          smtp.send_msg:
            - name: 'This is a server warning message'
            - profile: my-smtp-account
            - subject: 'Message from Salt'
            - recipient: admin@example.com
            - sender: admin@example.com
            - use_ssl: True

    name
        The message to send via SMTP
    '''
    ret = {'name': name,
           'changes': {},
           'result': None,
           'comment': ''}

    if profile is None and sender is None:
        ret['result'] = False
        ret['comment'] = 'Missing parameter sender or profile for state smtp.send_msg'
        return ret

    if __opts__['test']:
        ret['comment'] = 'Need to send message to {0}: {1}'.format(
            recipient,
            name,
        )
        return ret
    command = __salt__['smtp.send_msg'](
        message=name,
        recipient=recipient,
        profile=profile,
        subject=subject,
        sender=sender,
        use_ssl=use_ssl,
    )

    if command:
        ret['result'] = True
        ret['comment'] = 'Sent message to {0}: {1}'.format(recipient, name)
    else:
        ret['result'] = False
        ret['comment'] = 'Unable to send message to {0}: {1}'.format(recipient, name)
    return ret



# -*- coding: utf-8 -*-
'''
Beacon to emit Telegram messages

Requires the python-telegram-bot library

'''

# Import Python libs
from __future__ import absolute_import, unicode_literals
import logging
from salt.ext.six.moves import map

# Import 3rd Party libs
try:
    import telegram
    logging.getLogger('telegram').setLevel(logging.CRITICAL)
    HAS_TELEGRAM = True
except ImportError:
    HAS_TELEGRAM = False

log = logging.getLogger(__name__)


__virtualname__ = 'telegram_bot_msg'


def __virtual__():
    if HAS_TELEGRAM:
        return __virtualname__
    else:
        return False


def validate(config):
    '''
    Validate the beacon configuration
    '''
    if not isinstance(config, list):
        return False, ('Configuration for telegram_bot_msg '
                       'beacon must be a list.')

    _config = {}
    list(map(_config.update, config))

    if not all(_config.get(required_config)
               for required_config in ['token', 'accept_from']):
        return False, ('Not all required configuration for '
                       'telegram_bot_msg are set.')

    if not isinstance(_config.get('accept_from'), list):
        return False, ('Configuration for telegram_bot_msg, '
                       'accept_from must be a list of usernames.')

    return True, 'Valid beacon configuration.'


def beacon(config):
    '''
    Emit a dict with a key "msgs" whose value is a list of messages
    sent to the configured bot by one of the allowed usernames.

    .. code-block:: yaml

        beacons:
          telegram_bot_msg:
            - token: "<bot access token>"
            - accept_from:
              - "<valid username>"
            - interval: 10

    '''

    _config = {}
    list(map(_config.update, config))

    log.debug('telegram_bot_msg beacon starting')
    ret = []
    output = {}
    output['msgs'] = []

    bot = telegram.Bot(_config['token'])
    updates = bot.get_updates(limit=100, timeout=0, network_delay=10)

    log.debug('Num updates: %d', len(updates))
    if not updates:
        log.debug('Telegram Bot beacon has no new messages')
        return ret

    latest_update_id = 0
    for update in updates:
        message = update.message

        if update.update_id > latest_update_id:
            latest_update_id = update.update_id

        if message.chat.username in _config['accept_from']:
            output['msgs'].append(message.to_dict())

    # mark in the server that previous messages are processed
    bot.get_updates(offset=latest_update_id + 1)

    log.debug('Emitting %d messages.', len(output['msgs']))
    if output['msgs']:
        ret.append(output)
    return ret



# -*- coding: utf-8 -*-
'''
Connection module for Amazon APIGateway

.. versionadded:: 2016.11.0

:configuration: This module accepts explicit Lambda credentials but can also
    utilize IAM roles assigned to the instance trough Instance Profiles.
    Dynamic credentials are then automatically obtained from AWS API and no
    further configuration is necessary. More Information available at:

    .. code-block:: text

        http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/iam-roles-for-amazon-ec2.html

    If IAM roles are not used you need to specify them either in a pillar or
    in the minion's config file:

    .. code-block:: yaml

        apigateway.keyid: GKTADJGHEIQSXMKKRBJ08H
        apigateway.key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs

    A region may also be specified in the configuration:

    .. code-block:: yaml

        apigateway.region: us-west-2

    If a region is not specified, the default is us-east-1.

    It's also possible to specify key, keyid and region via a profile, either
    as a passed in dict, or as a string to pull from pillars or minion config:

    .. code-block:: yaml

        myprofile:
            keyid: GKTADJGHEIQSXMKKRBJ08H
            key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
            region: us-west-2

.. versionchanged:: 2015.8.0
    All methods now return a dictionary. Create and delete methods return:

    .. code-block:: yaml

        created: true

    or

    .. code-block:: yaml

        created: false
        error:
          message: error message

    Request methods (e.g., `describe_apigateway`) return:

    .. code-block:: yaml

        apigateway:
          - {...}
          - {...}

    or

    .. code-block:: yaml

        error:
          message: error message

:depends: boto3

'''
# keep lint from choking on _get_conn and _cache_id
# pylint: disable=E0602

# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals
import logging
import datetime

# Import Salt libs
from salt.ext import six
import salt.utils.compat
import salt.utils.json
import salt.utils.versions

log = logging.getLogger(__name__)

# Import third party libs

# pylint: disable=import-error
try:
    # pylint: disable=unused-import
    import boto
    import boto3
    # pylint: enable=unused-import
    from botocore.exceptions import ClientError
    from botocore import __version__ as found_botocore_version
    logging.getLogger('boto').setLevel(logging.CRITICAL)
    logging.getLogger('boto3').setLevel(logging.CRITICAL)
    HAS_BOTO = True
except ImportError:
    HAS_BOTO = False
# pylint: enable=import-error


def __virtual__():
    '''
    Only load if boto libraries exist and if boto libraries are greater than
    a given version.
    '''
    # the boto_apigateway execution module relies on the connect_to_region() method
    # which was added in boto 2.8.0
    # https://github.com/boto/boto/commit/33ac26b416fbb48a60602542b4ce15dcc7029f12
    return salt.utils.versions.check_boto_reqs(
        boto_ver='2.8.0',
        boto3_ver='1.2.1',
        botocore_ver='1.4.49'
    )


def __init__(opts):
    salt.utils.compat.pack_dunder(__name__)
    if HAS_BOTO:
        __utils__['boto3.assign_funcs'](__name__, 'apigateway')


def _convert_datetime_str(response):
    '''
    modify any key-value pair where value is a datetime object to a string.
    '''
    if response:
        return dict([(k, '{0}'.format(v)) if isinstance(v, datetime.date) else (k, v) for k, v in six.iteritems(response)])
    return None


def _filter_apis(name, apis):
    '''
    Return list of api items matching the given name.
    '''
    return [api for api in apis if api['name'] == name]


def _filter_apis_desc(desc, apis):
    '''
    Return list of api items matching the given description.
    '''
    return [api for api in apis if api['description'] == desc]


def _multi_call(function, contentkey, *args, **kwargs):
    '''
    Retrieve full list of values for the contentkey from a boto3 ApiGateway
    client function that may be paged via 'position'
    '''
    ret = function(*args, **kwargs)
    position = ret.get('position')

    while position:
        more = function(*args, position=position, **kwargs)
        ret[contentkey].extend(more[contentkey])
        position = more.get('position')
    return ret.get(contentkey)


def _find_apis_by_name(name, description=None,
                       region=None, key=None, keyid=None, profile=None):

    '''
    get and return list of matching rest api information by the given name and desc.
    If rest api name evaluates to False, return all apis w/o filtering the name.
    '''
    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        apis = _multi_call(conn.get_rest_apis, 'items')
        if name:
            apis = _filter_apis(name, apis)
        if description is not None:
            apis = _filter_apis_desc(description, apis)
        return {'restapi': [_convert_datetime_str(api) for api in apis]}
    except ClientError as e:
        return {'error': __utils__['boto3.get_error'](e)}


def describe_apis(name=None, description=None, region=None, key=None, keyid=None, profile=None):

    '''
    Returns all rest apis in the defined region.  If optional parameter name is included,
    returns all rest apis matching the name in the defined region.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.describe_apis

        salt myminion boto_apigateway.describe_apis name='api name'

        salt myminion boto_apigateway.describe_apis name='api name' description='desc str'

    '''

    if name:
        return _find_apis_by_name(name, description=description,
                                  region=region, key=key, keyid=keyid, profile=profile)
    else:
        return _find_apis_by_name('', description=description,
                                  region=region, key=key, keyid=keyid, profile=profile)


def api_exists(name, description=None, region=None, key=None, keyid=None, profile=None):
    '''
    Check to see if the given Rest API Name and optionlly description exists.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.exists myapi_name

    '''
    apis = _find_apis_by_name(name, description=description,
                              region=region, key=key, keyid=keyid, profile=profile)
    return {'exists': bool(apis.get('restapi'))}


def create_api(name, description, cloneFrom=None,
               region=None, key=None, keyid=None, profile=None):
    '''
    Create a new REST API Service with the given name

    Returns {created: True} if the rest api was created and returns
    {created: False} if the rest api was not created.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.create_api myapi_name api_description

    '''
    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        if cloneFrom:
            api = conn.create_rest_api(name=name, description=description, cloneFrom=cloneFrom)
        else:
            api = conn.create_rest_api(name=name, description=description)
        api = _convert_datetime_str(api)
        return {'created': True, 'restapi': api} if api else {'created': False}
    except ClientError as e:
        return {'created': False, 'error': __utils__['boto3.get_error'](e)}


def delete_api(name, description=None, region=None, key=None, keyid=None, profile=None):
    '''
    Delete all REST API Service with the given name and an optional API description

    Returns {deleted: True, count: deleted_count} if apis were deleted, and
    returns {deleted: False} if error or not found.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.delete_api myapi_name

        salt myminion boto_apigateway.delete_api myapi_name description='api description'

    '''
    try:
        conn_params = dict(region=region, key=key, keyid=keyid, profile=profile)
        r = _find_apis_by_name(name, description=description, **conn_params)
        apis = r.get('restapi')
        if apis:
            conn = _get_conn(**conn_params)
            for api in apis:
                conn.delete_rest_api(restApiId=api['id'])
            return {'deleted': True, 'count': len(apis)}
        else:
            return {'deleted': False}
    except ClientError as e:
        return {'deleted': False, 'error': __utils__['boto3.get_error'](e)}


def describe_api_resources(restApiId, region=None, key=None, keyid=None, profile=None):
    '''
    Given rest api id, return all resources for this api.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.describe_api_resources myapi_id

    '''
    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        resources = sorted(_multi_call(conn.get_resources, 'items', restApiId=restApiId),
                           key=lambda k: k['path'])

        return {'resources': resources}
    except ClientError as e:
        return {'error': __utils__['boto3.get_error'](e)}


def describe_api_resource(restApiId, path,
                          region=None, key=None, keyid=None, profile=None):
    '''
    Given rest api id, and an absolute resource path, returns the resource id for
    the given path.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.describe_api_resource myapi_id resource_path

    '''
    r = describe_api_resources(restApiId, region=region, key=key, keyid=keyid, profile=profile)
    resources = r.get('resources')
    if resources is None:
        return r
    for resource in resources:
        if resource['path'] == path:
            return {'resource': resource}
    return {'resource': None}


def create_api_resources(restApiId, path,
                         region=None, key=None, keyid=None, profile=None):
    '''
    Given rest api id, and an absolute resource path, create all the resources and
    return all resources in the resourcepath, returns False on failure.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.create_api_resources myapi_id resource_path

    '''
    path_parts = path.split('/')
    created = []
    current_path = ''
    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        for path_part in path_parts:
            if current_path == '/':
                current_path = '{0}{1}'.format(current_path, path_part)
            else:
                current_path = '{0}/{1}'.format(current_path, path_part)
            r = describe_api_resource(restApiId, current_path,
                                      region=region, key=key, keyid=keyid, profile=profile)
            resource = r.get('resource')
            if not resource:
                resource = conn.create_resource(restApiId=restApiId, parentId=created[-1]['id'], pathPart=path_part)
            created.append(resource)

        if created:
            return {'created': True, 'restApiId': restApiId, 'resources': created}
        else:
            return {'created': False, 'error': 'unexpected error.'}
    except ClientError as e:
        return {'created': False, 'error': __utils__['boto3.get_error'](e)}


def delete_api_resources(restApiId, path,
                         region=None, key=None, keyid=None, profile=None):
    '''
    Given restApiId and an absolute resource path, delete the resources starting
    from the absoluate resource path.  If resourcepath is the root resource '/',
    the function will return False.  Returns False on failure.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.delete_api_resources myapi_id, resource_path

    '''
    if path == '/':
        return {'deleted': False, 'error': 'use delete_api to remove the root resource'}
    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        r = describe_api_resource(restApiId, path, region=region, key=key, keyid=keyid, profile=profile)
        resource = r.get('resource')
        if resource:
            conn.delete_resource(restApiId=restApiId, resourceId=resource['id'])
            return {'deleted': True}
        else:
            return {'deleted': False, 'error': 'no resource found by {0}'.format(path)}
    except ClientError as e:
        return {'created': False, 'error': __utils__['boto3.get_error'](e)}


def describe_api_resource_method(restApiId, resourcePath, httpMethod,
                                 region=None, key=None, keyid=None, profile=None):
    '''
    Given rest api id, resource path, and http method (must be one of DELETE,
    GET, HEAD, OPTIONS, PATCH, POST, PUT), return the method for the
    api/resource path if defined.  Return False if method is not defined.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.describe_api_resource_method myapi_id resource_path httpmethod

    '''
    r = describe_api_resource(restApiId, resourcePath,
                              region=region, key=key, keyid=keyid, profile=profile)
    resource = r.get('resource')
    if not resource:
        return {'error': 'no such resource'}

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        method = conn.get_method(restApiId=restApiId, resourceId=resource['id'], httpMethod=httpMethod)
        return {'method': method}
    except ClientError as e:
        return {'error': __utils__['boto3.get_error'](e)}


def describe_api_key(apiKey, region=None, key=None, keyid=None, profile=None):
    '''
    Gets info about the given api key

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.describe_api_key apigw_api_key

    '''
    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        response = conn.get_api_key(apiKey=apiKey)
        return {'apiKey': _convert_datetime_str(response)}
    except ClientError as e:
        return {'error': __utils__['boto3.get_error'](e)}


def describe_api_keys(region=None, key=None, keyid=None, profile=None):
    '''
    Gets information about the defined API Keys.  Return list of apiKeys.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.describe_api_keys

    '''
    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        apikeys = _multi_call(conn.get_api_keys, 'items')

        return {'apiKeys': [_convert_datetime_str(apikey) for apikey in apikeys]}
    except ClientError as e:
        return {'error': __utils__['boto3.get_error'](e)}


def create_api_key(name, description, enabled=True, stageKeys=None,
                   region=None, key=None, keyid=None, profile=None):
    '''
    Create an API key given name and description.

    An optional enabled argument can be provided.  If provided, the
    valid values are True|False.  This argument defaults to True.

    An optional stageKeys argument can be provided in the form of
    list of dictionary with 'restApiId' and 'stageName' as keys.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.create_api_key name description

        salt myminion boto_apigateway.create_api_key name description enabled=False

        salt myminion boto_apigateway.create_api_key name description \\
             stageKeys='[{"restApiId": "id", "stageName": "stagename"}]'

    '''

    try:
        stageKeys = list() if stageKeys is None else stageKeys

        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        response = conn.create_api_key(name=name, description=description,
                                       enabled=enabled, stageKeys=stageKeys)
        if not response:
            return {'created': False}

        return {'created': True, 'apiKey': _convert_datetime_str(response)}
    except ClientError as e:
        return {'created': False, 'error': __utils__['boto3.get_error'](e)}


def delete_api_key(apiKey, region=None, key=None, keyid=None, profile=None):
    '''
    Deletes a given apiKey

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.delete_api_key apikeystring

    '''
    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        conn.delete_api_key(apiKey=apiKey)
        return {'deleted': True}
    except ClientError as e:
        return {'deleted': False, 'error': __utils__['boto3.get_error'](e)}


def _api_key_patch_replace(conn, apiKey, path, value):
    '''
    the replace patch operation on an ApiKey resource
    '''
    response = conn.update_api_key(apiKey=apiKey,
                                   patchOperations=[{'op': 'replace', 'path': path, 'value': value}])
    return response


def _api_key_patchops(op, pvlist):
    '''
    helper function to return patchOperations object
    '''
    return [{'op': op, 'path': p, 'value': v} for (p, v) in pvlist]


def _api_key_patch_add(conn, apiKey, pvlist):
    '''
    the add patch operation for a list of (path, value) tuples on an ApiKey resource list path
    '''
    response = conn.update_api_key(apiKey=apiKey,
                                   patchOperations=_api_key_patchops('add', pvlist))
    return response


def _api_key_patch_remove(conn, apiKey, pvlist):
    '''
    the remove patch operation for a list of (path, value) tuples on an ApiKey resource list path
    '''
    response = conn.update_api_key(apiKey=apiKey,
                                   patchOperations=_api_key_patchops('remove', pvlist))
    return response


def update_api_key_description(apiKey, description, region=None, key=None, keyid=None, profile=None):
    '''
    update the given apiKey with the given description.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.update_api_key_description api_key description

    '''
    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        response = _api_key_patch_replace(conn, apiKey, '/description', description)
        return {'updated': True, 'apiKey': _convert_datetime_str(response)}
    except ClientError as e:
        return {'updated': False, 'error': __utils__['boto3.get_error'](e)}


def enable_api_key(apiKey, region=None, key=None, keyid=None, profile=None):
    '''
    enable the given apiKey.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.enable_api_key api_key

    '''
    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        response = _api_key_patch_replace(conn, apiKey, '/enabled', 'True')
        return {'apiKey': _convert_datetime_str(response)}
    except ClientError as e:
        return {'error': __utils__['boto3.get_error'](e)}


def disable_api_key(apiKey, region=None, key=None, keyid=None, profile=None):
    '''
    disable the given apiKey.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.enable_api_key api_key

    '''
    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        response = _api_key_patch_replace(conn, apiKey, '/enabled', 'False')
        return {'apiKey': _convert_datetime_str(response)}
    except ClientError as e:
        return {'error': __utils__['boto3.get_error'](e)}


def associate_api_key_stagekeys(apiKey, stagekeyslist, region=None, key=None, keyid=None, profile=None):
    '''
    associate the given stagekeyslist to the given apiKey.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.associate_stagekeys_api_key \\
                api_key '["restapi id/stage name", ...]'

    '''
    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        pvlist = [('/stages', stagekey) for stagekey in stagekeyslist]
        response = _api_key_patch_add(conn, apiKey, pvlist)
        return {'associated': True, 'apiKey': _convert_datetime_str(response)}
    except ClientError as e:
        return {'associated': False, 'error': __utils__['boto3.get_error'](e)}


def disassociate_api_key_stagekeys(apiKey, stagekeyslist, region=None, key=None, keyid=None, profile=None):
    '''
    disassociate the given stagekeyslist to the given apiKey.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.disassociate_stagekeys_api_key \\
                api_key '["restapi id/stage name", ...]'

    '''
    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        pvlist = [('/stages', stagekey) for stagekey in stagekeyslist]
        response = _api_key_patch_remove(conn, apiKey, pvlist)
        return {'disassociated': True}
    except ClientError as e:
        return {'disassociated': False, 'error': __utils__['boto3.get_error'](e)}


def describe_api_deployments(restApiId, region=None, key=None, keyid=None, profile=None):
    '''
    Gets information about the defined API Deployments.  Return list of api deployments.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.describe_api_deployments restApiId

    '''
    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        deployments = []
        _deployments = conn.get_deployments(restApiId=restApiId)

        while True:
            if _deployments:
                deployments = deployments + _deployments['items']
                if 'position' not in _deployments:
                    break
                _deployments = conn.get_deployments(restApiId=restApiId, position=_deployments['position'])

        return {'deployments': [_convert_datetime_str(deployment) for deployment in deployments]}
    except ClientError as e:
        return {'error': __utils__['boto3.get_error'](e)}


def describe_api_deployment(restApiId, deploymentId, region=None, key=None, keyid=None, profile=None):
    '''
    Get API deployment for a given restApiId and deploymentId.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.describe_api_deployent restApiId deploymentId

    '''
    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        deployment = conn.get_deployment(restApiId=restApiId, deploymentId=deploymentId)
        return {'deployment': _convert_datetime_str(deployment)}
    except ClientError as e:
        return {'error': __utils__['boto3.get_error'](e)}


def activate_api_deployment(restApiId, stageName, deploymentId,
                            region=None, key=None, keyid=None, profile=None):
    '''
    Activates previously deployed deployment for a given stage

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.activate_api_deployent restApiId stagename deploymentId

    '''
    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        response = conn.update_stage(restApiId=restApiId, stageName=stageName,
                                     patchOperations=[{'op': 'replace',
                                                       'path': '/deploymentId',
                                                       'value': deploymentId}])
        return {'set': True, 'response': _convert_datetime_str(response)}
    except ClientError as e:
        return {'set': False, 'error': __utils__['boto3.get_error'](e)}


def create_api_deployment(restApiId, stageName, stageDescription='', description='', cacheClusterEnabled=False,
                          cacheClusterSize='0.5', variables=None,
                          region=None, key=None, keyid=None, profile=None):
    '''
    Creates a new API deployment.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.create_api_deployent restApiId stagename stageDescription='' \\
        description='' cacheClusterEnabled=True|False cacheClusterSize=0.5 variables='{"name": "value"}'

    '''
    try:
        variables = dict() if variables is None else variables

        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        deployment = conn.create_deployment(restApiId=restApiId, stageName=stageName,
                                            stageDescription=stageDescription, description=description,
                                            cacheClusterEnabled=cacheClusterEnabled, cacheClusterSize=cacheClusterSize,
                                            variables=variables)
        return {'created': True, 'deployment': _convert_datetime_str(deployment)}
    except ClientError as e:
        return {'created': False, 'error': __utils__['boto3.get_error'](e)}


def delete_api_deployment(restApiId, deploymentId, region=None, key=None, keyid=None, profile=None):
    '''
    Deletes API deployment for a given restApiId and deploymentID

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.delete_api_deployent restApiId deploymentId

    '''
    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        conn.delete_deployment(restApiId=restApiId, deploymentId=deploymentId)
        return {'deleted': True}
    except ClientError as e:
        return {'deleted': False, 'error': __utils__['boto3.get_error'](e)}


def overwrite_api_stage_variables(restApiId, stageName, variables, region=None, key=None, keyid=None, profile=None):
    '''
    Overwrite the stage variables for the given restApiId and stage name with the given variables,
    variables must be in the form of a dictionary.  Overwrite will always remove all the existing
    stage variables associated with the given restApiId and stage name, follow by the adding of all the
    variables specified in the variables dictionary

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.overwrite_api_stage_variables restApiId stageName variables='{"name": "value"}'

    '''
    try:
        res = describe_api_stage(restApiId, stageName, region=region, key=key, keyid=keyid, profile=profile)
        if res.get('error'):
            return {'overwrite': False, 'error': res.get('error')}

        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

        # remove all existing variables that are not in the given variables,
        # followed by adding of the variables
        stage = res.get('stage')
        old_vars = stage.get('variables', {})
        patch_ops = []
        for old_var in old_vars:
            if old_var not in variables:
                patch_ops.append(dict(op='remove',
                                      path='/variables/{0}'.format(old_var),
                                      value=''))
        for var, val in six.iteritems(variables):
            if var not in old_vars or old_vars[var] != val:
                patch_ops.append(dict(op='replace',
                                      path='/variables/{0}'.format(var),
                                      value=val))

        if patch_ops:
            stage = conn.update_stage(restApiId=restApiId, stageName=stageName,
                                      patchOperations=patch_ops)

        return {'overwrite': True, 'stage': _convert_datetime_str(stage)}
    except ClientError as e:
        return {'overwrite': False, 'error': __utils__['boto3.get_error'](e)}


def describe_api_stage(restApiId, stageName, region=None, key=None, keyid=None, profile=None):
    '''
    Get API stage for a given apiID and stage name

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.describe_api_stage restApiId stageName

    '''
    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        stage = conn.get_stage(restApiId=restApiId, stageName=stageName)
        return {'stage': _convert_datetime_str(stage)}
    except ClientError as e:
        return {'error': __utils__['boto3.get_error'](e)}


def describe_api_stages(restApiId, deploymentId, region=None, key=None, keyid=None, profile=None):
    '''
    Get all API stages for a given apiID and deploymentID

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.describe_api_stages restApiId deploymentId

    '''
    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        stages = conn.get_stages(restApiId=restApiId, deploymentId=deploymentId)
        return {'stages': [_convert_datetime_str(stage) for stage in stages['item']]}
    except ClientError as e:
        return {'error': __utils__['boto3.get_error'](e)}


def create_api_stage(restApiId, stageName, deploymentId, description='',
                     cacheClusterEnabled=False, cacheClusterSize='0.5', variables=None,
                     region=None, key=None, keyid=None, profile=None):
    '''
    Creates a new API stage for a given restApiId and deploymentId.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.create_api_stage restApiId stagename deploymentId \\
            description='' cacheClusterEnabled=True|False cacheClusterSize='0.5' variables='{"name": "value"}'

    '''
    try:
        variables = dict() if variables is None else variables

        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        stage = conn.create_stage(restApiId=restApiId, stageName=stageName, deploymentId=deploymentId,
                                  description=description, cacheClusterEnabled=cacheClusterEnabled,
                                  cacheClusterSize=cacheClusterSize, variables=variables)
        return {'created': True, 'stage': _convert_datetime_str(stage)}
    except ClientError as e:
        return {'created': False, 'error': __utils__['boto3.get_error'](e)}


def delete_api_stage(restApiId, stageName, region=None, key=None, keyid=None, profile=None):
    '''
    Deletes stage identified by stageName from API identified by restApiId

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.delete_api_stage restApiId stageName

    '''
    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        conn.delete_stage(restApiId=restApiId, stageName=stageName)
        return {'deleted': True}
    except ClientError as e:
        return {'deleted': False, 'error': __utils__['boto3.get_error'](e)}


def flush_api_stage_cache(restApiId, stageName, region=None, key=None, keyid=None, profile=None):
    '''
    Flushes cache for the stage identified by stageName from API identified by restApiId

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.flush_api_stage_cache restApiId stageName

    '''
    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        conn.flush_stage_cache(restApiId=restApiId, stageName=stageName)
        return {'flushed': True}
    except ClientError as e:
        return {'flushed': False, 'error': __utils__['boto3.get_error'](e)}


def create_api_method(restApiId, resourcePath, httpMethod, authorizationType,
                      apiKeyRequired=False, requestParameters=None, requestModels=None,
                      region=None, key=None, keyid=None, profile=None):
    '''
    Creates API method for a resource in the given API

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.create_api_method restApiId resourcePath, httpMethod, authorizationType, \\
            apiKeyRequired=False, requestParameters='{"name", "value"}', requestModels='{"content-type", "value"}'

    '''
    try:
        resource = describe_api_resource(restApiId, resourcePath, region=region,
                                         key=key, keyid=keyid, profile=profile).get('resource')
        if resource:
            requestParameters = dict() if requestParameters is None else requestParameters
            requestModels = dict() if requestModels is None else requestModels

            conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
            method = conn.put_method(restApiId=restApiId, resourceId=resource['id'], httpMethod=httpMethod,
                                     authorizationType=str(authorizationType), apiKeyRequired=apiKeyRequired,  # future lint: disable=blacklisted-function
                                     requestParameters=requestParameters, requestModels=requestModels)
            return {'created': True, 'method': method}
        return {'created': False, 'error': 'Failed to create method'}

    except ClientError as e:
        return {'created': False, 'error': __utils__['boto3.get_error'](e)}


def describe_api_method(restApiId, resourcePath, httpMethod, region=None, key=None, keyid=None, profile=None):
    '''
    Get API method for a resource in the given API

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.describe_api_method restApiId resourcePath httpMethod

    '''
    try:
        resource = describe_api_resource(restApiId, resourcePath, region=region,
                                         key=key, keyid=keyid, profile=profile).get('resource')
        if resource:
            conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
            method = conn.get_method(restApiId=restApiId, resourceId=resource['id'], httpMethod=httpMethod)
            return {'method': _convert_datetime_str(method)}
        return {'error': 'get API method failed: no such resource'}
    except ClientError as e:
        return {'error': __utils__['boto3.get_error'](e)}


def delete_api_method(restApiId, resourcePath, httpMethod, region=None, key=None, keyid=None, profile=None):
    '''
    Delete API method for a resource in the given API

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.delete_api_method restApiId resourcePath httpMethod

    '''
    try:
        resource = describe_api_resource(restApiId, resourcePath, region=region,
                                         key=key, keyid=keyid, profile=profile).get('resource')
        if resource:
            conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
            conn.delete_method(restApiId=restApiId, resourceId=resource['id'], httpMethod=httpMethod)
            return {'deleted': True}
        return {'deleted': False, 'error': 'get API method failed: no such resource'}
    except ClientError as e:
        return {'deleted': False, 'error': __utils__['boto3.get_error'](e)}


def create_api_method_response(restApiId, resourcePath, httpMethod, statusCode, responseParameters=None,
                               responseModels=None, region=None, key=None, keyid=None, profile=None):
    '''
    Create API method response for a method on a given resource in the given API

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.create_api_method_response restApiId resourcePath httpMethod \\
               statusCode responseParameters='{"name", "True|False"}' responseModels='{"content-type", "model"}'

    '''
    try:
        resource = describe_api_resource(restApiId, resourcePath, region=region,
                                         key=key, keyid=keyid, profile=profile).get('resource')
        if resource:
            responseParameters = dict() if responseParameters is None else responseParameters
            responseModels = dict() if responseModels is None else responseModels

            conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
            response = conn.put_method_response(restApiId=restApiId, resourceId=resource['id'],
                                                httpMethod=httpMethod, statusCode=str(statusCode),  # future lint: disable=blacklisted-function
                                                responseParameters=responseParameters, responseModels=responseModels)
            return {'created': True, 'response': response}
        return {'created': False, 'error': 'no such resource'}
    except ClientError as e:
        return {'created': False, 'error': __utils__['boto3.get_error'](e)}


def delete_api_method_response(restApiId, resourcePath, httpMethod, statusCode,
                               region=None, key=None, keyid=None, profile=None):
    '''
    Delete API method response for a resource in the given API

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.delete_api_method_response restApiId resourcePath httpMethod statusCode

    '''
    try:
        resource = describe_api_resource(restApiId, resourcePath, region=region,
                                         key=key, keyid=keyid, profile=profile).get('resource')
        if resource:
            conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
            conn.delete_method_response(restApiId=restApiId, resourceId=resource['id'],
                                        httpMethod=httpMethod, statusCode=str(statusCode))  # future lint: disable=blacklisted-function
            return {'deleted': True}
        return {'deleted': False, 'error': 'no such resource'}
    except ClientError as e:
        return {'deleted': False, 'error': __utils__['boto3.get_error'](e)}


def describe_api_method_response(restApiId, resourcePath, httpMethod, statusCode,
                                 region=None, key=None, keyid=None, profile=None):
    '''
    Get API method response for a resource in the given API

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.describe_api_method_response restApiId resourcePath httpMethod statusCode

    '''
    try:
        resource = describe_api_resource(restApiId, resourcePath, region=region,
                                         key=key, keyid=keyid, profile=profile).get('resource')
        if resource:
            conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
            response = conn.get_method_response(restApiId=restApiId, resourceId=resource['id'],
                                                httpMethod=httpMethod, statusCode=str(statusCode))  # future lint: disable=blacklisted-function
            return {'response': _convert_datetime_str(response)}
        return {'error': 'no such resource'}
    except ClientError as e:
        return {'error': __utils__['boto3.get_error'](e)}


def describe_api_models(restApiId, region=None, key=None, keyid=None, profile=None):
    '''
    Get all models for a given API

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.describe_api_models restApiId

    '''
    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        models = _multi_call(conn.get_models, 'items', restApiId=restApiId)
        return {'models': [_convert_datetime_str(model) for model in models]}
    except ClientError as e:
        return {'error': __utils__['boto3.get_error'](e)}


def describe_api_model(restApiId, modelName, flatten=True, region=None, key=None, keyid=None, profile=None):
    '''
    Get a model by name for a given API

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.describe_api_model restApiId modelName [True]

    '''
    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        model = conn.get_model(restApiId=restApiId, modelName=modelName, flatten=flatten)
        return {'model': _convert_datetime_str(model)}
    except ClientError as e:
        return {'error': __utils__['boto3.get_error'](e)}


def api_model_exists(restApiId, modelName, region=None, key=None, keyid=None, profile=None):
    '''
    Check to see if the given modelName exists in the given restApiId

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.api_model_exists restApiId modelName
    '''
    r = describe_api_model(restApiId, modelName, region=region, key=key, keyid=keyid, profile=profile)

    return {'exists': bool(r.get('model'))}


def _api_model_patch_replace(conn, restApiId, modelName, path, value):
    '''
    the replace patch operation on a Model resource
    '''
    response = conn.update_model(restApiId=restApiId, modelName=modelName,
                                 patchOperations=[{'op': 'replace', 'path': path, 'value': value}])
    return response


def update_api_model_schema(restApiId, modelName, schema, region=None, key=None, keyid=None, profile=None):
    '''
    update the schema (in python dictionary format) for the given model in the given restApiId

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.update_api_model_schema restApiId modelName schema

    '''
    try:
        schema_json = salt.utils.json.dumps(schema) if isinstance(schema, dict) else schema
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        response = _api_model_patch_replace(conn, restApiId, modelName, '/schema', schema_json)
        return {'updated': True, 'model': _convert_datetime_str(response)}
    except ClientError as e:
        return {'updated': False, 'error': __utils__['boto3.get_error'](e)}


def delete_api_model(restApiId, modelName, region=None, key=None, keyid=None, profile=None):
    '''
    Delete a model identified by name in a given API

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.delete_api_model restApiId modelName

    '''
    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        conn.delete_model(restApiId=restApiId, modelName=modelName)
        return {'deleted': True}
    except ClientError as e:
        return {'deleted': False, 'error': __utils__['boto3.get_error'](e)}


def create_api_model(restApiId, modelName, modelDescription, schema, contentType='application/json',
                     region=None, key=None, keyid=None, profile=None):
    '''
    Create a new model in a given API with a given schema, currently only contentType supported is
    'application/json'

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.create_api_model restApiId modelName modelDescription '<schema>' 'content-type'

    '''
    try:
        schema_json = salt.utils.json.dumps(schema) if isinstance(schema, dict) else schema
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        model = conn.create_model(restApiId=restApiId, name=modelName, description=modelDescription,
                                  schema=schema_json, contentType=contentType)
        return {'created': True, 'model': _convert_datetime_str(model)}
    except ClientError as e:
        return {'created': False, 'error': __utils__['boto3.get_error'](e)}


def describe_api_integration(restApiId, resourcePath, httpMethod, region=None, key=None, keyid=None, profile=None):
    '''
    Get an integration for a given method in a given API

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.describe_api_integration restApiId resourcePath httpMethod

    '''
    try:
        resource = describe_api_resource(restApiId, resourcePath, region=region,
                                         key=key, keyid=keyid, profile=profile).get('resource')
        if resource:
            conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
            integration = conn.get_integration(restApiId=restApiId, resourceId=resource['id'], httpMethod=httpMethod)
            return {'integration': _convert_datetime_str(integration)}
        return {'error': 'no such resource'}
    except ClientError as e:
        return {'error': __utils__['boto3.get_error'](e)}


def describe_api_integration_response(restApiId, resourcePath, httpMethod, statusCode,
                                      region=None, key=None, keyid=None, profile=None):
    '''
    Get an integration response for a given method in a given API

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.describe_api_integration_response restApiId resourcePath httpMethod statusCode

    '''
    try:
        resource = describe_api_resource(restApiId, resourcePath, region=region,
                                         key=key, keyid=keyid, profile=profile).get('resource')
        if resource:
            conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
            response = conn.get_integration_response(restApiId=restApiId, resourceId=resource['id'],
                                                     httpMethod=httpMethod, statusCode=statusCode)
            return {'response': _convert_datetime_str(response)}
        return {'error': 'no such resource'}
    except ClientError as e:
        return {'error': __utils__['boto3.get_error'](e)}


def delete_api_integration(restApiId, resourcePath, httpMethod, region=None, key=None, keyid=None, profile=None):
    '''
    Deletes an integration for a given method in a given API

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.delete_api_integration restApiId resourcePath httpMethod

    '''
    try:
        resource = describe_api_resource(restApiId, resourcePath, region=region,
                                         key=key, keyid=keyid, profile=profile).get('resource')
        if resource:
            conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
            conn.delete_integration(restApiId=restApiId, resourceId=resource['id'], httpMethod=httpMethod)
            return {'deleted': True}
        return {'deleted': False, 'error': 'no such resource'}
    except ClientError as e:
        return {'deleted': False, 'error': __utils__['boto3.get_error'](e)}


def delete_api_integration_response(restApiId, resourcePath, httpMethod, statusCode,
                                    region=None, key=None, keyid=None, profile=None):
    '''
    Deletes an integration response for a given method in a given API

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.delete_api_integration_response restApiId resourcePath httpMethod statusCode

    '''
    try:
        resource = describe_api_resource(restApiId, resourcePath, region=region,
                                         key=key, keyid=keyid, profile=profile).get('resource')
        if resource:
            conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
            conn.delete_integration_response(restApiId=restApiId, resourceId=resource['id'],
                                             httpMethod=httpMethod, statusCode=statusCode)
            return {'deleted': True}
        return {'deleted': False, 'error': 'no such resource'}
    except ClientError as e:
        return {'deleted': False, 'error': __utils__['boto3.get_error'](e)}


def _get_role_arn(name, region=None, key=None, keyid=None, profile=None):
    '''
    Helper function to get an ARN if name does not look like an ARN.
    '''
    if name.startswith('arn:aws:iam:'):
        return name

    account_id = __salt__['boto_iam.get_account_id'](
        region=region, key=key, keyid=keyid, profile=profile
    )

    return 'arn:aws:iam::{0}:role/{1}'.format(account_id, name)


def create_api_integration(restApiId, resourcePath, httpMethod, integrationType, integrationHttpMethod,
                           uri, credentials, requestParameters=None, requestTemplates=None,
                           region=None, key=None, keyid=None, profile=None):
    '''
    Creates an integration for a given method in a given API.
    If integrationType is MOCK, uri and credential parameters will be ignored.

    uri is in the form of (substitute APIGATEWAY_REGION and LAMBDA_FUNC_ARN)
    "arn:aws:apigateway:APIGATEWAY_REGION:lambda:path/2015-03-31/functions/LAMBDA_FUNC_ARN/invocations"

    credentials is in the form of an iam role name or role arn.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.create_api_integration restApiId resourcePath httpMethod \\
                             integrationType integrationHttpMethod uri credentials ['{}' ['{}']]

    '''
    try:
        credentials = _get_role_arn(credentials, region=region, key=key, keyid=keyid, profile=profile)
        resource = describe_api_resource(restApiId, resourcePath, region=region,
                                         key=key, keyid=keyid, profile=profile).get('resource')
        if resource:
            requestParameters = dict() if requestParameters is None else requestParameters
            requestTemplates = dict() if requestTemplates is None else requestTemplates

            conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
            if httpMethod.lower() == 'options':
                uri = ""
                credentials = ""

            integration = conn.put_integration(restApiId=restApiId, resourceId=resource['id'], httpMethod=httpMethod,
                                               type=integrationType, integrationHttpMethod=integrationHttpMethod,
                                               uri=uri, credentials=credentials, requestParameters=requestParameters,
                                               requestTemplates=requestTemplates)
            return {'created': True, 'integration': integration}
        return {'created': False, 'error': 'no such resource'}
    except ClientError as e:
        return {'created': False, 'error': __utils__['boto3.get_error'](e)}


def create_api_integration_response(restApiId, resourcePath, httpMethod, statusCode, selectionPattern,
                                    responseParameters=None, responseTemplates=None,
                                    region=None, key=None, keyid=None, profile=None):
    '''
    Creates an integration response for a given method in a given API

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.create_api_integration_response restApiId resourcePath httpMethod \\
                            statusCode selectionPattern ['{}' ['{}']]

    '''
    try:
        resource = describe_api_resource(restApiId, resourcePath, region=region,
                                         key=key, keyid=keyid, profile=profile).get('resource')
        if resource:
            responseParameters = dict() if responseParameters is None else responseParameters
            responseTemplates = dict() if responseTemplates is None else responseTemplates

            conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
            response = conn.put_integration_response(restApiId=restApiId, resourceId=resource['id'],
                                                     httpMethod=httpMethod, statusCode=statusCode,
                                                     selectionPattern=selectionPattern,
                                                     responseParameters=responseParameters,
                                                     responseTemplates=responseTemplates)
            return {'created': True, 'response': response}
        return {'created': False, 'error': 'no such resource'}
    except ClientError as e:
        return {'created': False, 'error': __utils__['boto3.get_error'](e)}


def _filter_plans(attr, name, plans):
    '''
    Helper to return list of usage plan items matching the given attribute value.
    '''
    return [plan for plan in plans if plan[attr] == name]


def describe_usage_plans(name=None, plan_id=None, region=None, key=None, keyid=None, profile=None):
    '''
    Returns a list of existing usage plans, optionally filtered to match a given plan name

    .. versionadded:: 2017.7.0

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.describe_usage_plans
        salt myminion boto_apigateway.describe_usage_plans name='usage plan name'
        salt myminion boto_apigateway.describe_usage_plans plan_id='usage plan id'

    '''
    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        plans = _multi_call(conn.get_usage_plans, 'items')
        if name:
            plans = _filter_plans('name', name, plans)
        if plan_id:
            plans = _filter_plans('id', plan_id, plans)

        return {'plans': [_convert_datetime_str(plan) for plan in plans]}

    except ClientError as e:
        return {'error': __utils__['boto3.get_error'](e)}


def _validate_throttle(throttle):
    '''
    Helper to verify that throttling parameters are valid
    '''
    if throttle is not None:
        if not isinstance(throttle, dict):
            raise TypeError('throttle must be a dictionary, provided value: {0}'.format(throttle))


def _validate_quota(quota):
    '''
    Helper to verify that quota parameters are valid
    '''
    if quota is not None:
        if not isinstance(quota, dict):
            raise TypeError('quota must be a dictionary, provided value: {0}'.format(quota))
        periods = ['DAY', 'WEEK', 'MONTH']
        if 'period' not in quota or quota['period'] not in periods:
            raise ValueError('quota must have a valid period specified, valid values are {0}'.format(','.join(periods)))
        if 'limit' not in quota:
            raise ValueError('quota limit must have a valid value')


def create_usage_plan(name, description=None, throttle=None, quota=None, region=None, key=None, keyid=None, profile=None):
    '''
    Creates a new usage plan with throttling and quotas optionally applied

    .. versionadded:: 2017.7.0

    name
        Name of the usage plan

    throttle
        A dictionary consisting of the following keys:

        rateLimit
            requests per second at steady rate, float

        burstLimit
            maximum number of requests per second, integer

    quota
        A dictionary consisting of the following keys:

        limit
            number of allowed requests per specified quota period [required if quota parameter is present]

        offset
            number of requests to be subtracted from limit at the beginning of the period [optional]

        period
            quota period, must be one of DAY, WEEK, or MONTH. [required if quota parameter is present

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.create_usage_plan name='usage plan name' throttle='{"rateLimit": 10.0, "burstLimit": 10}'

    '''
    try:
        _validate_throttle(throttle)
        _validate_quota(quota)

        values = dict(name=name)
        if description:
            values['description'] = description
        if throttle:
            values['throttle'] = throttle
        if quota:
            values['quota'] = quota

        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        res = conn.create_usage_plan(**values)
        return {'created': True, 'result': res}
    except ClientError as e:
        return {'error': __utils__['boto3.get_error'](e)}
    except (TypeError, ValueError) as e:
        return {'error': six.text_type(e)}


def update_usage_plan(plan_id, throttle=None, quota=None, region=None, key=None, keyid=None, profile=None):
    '''
    Updates an existing usage plan with throttling and quotas

    .. versionadded:: 2017.7.0

    plan_id
        Id of the created usage plan

    throttle
        A dictionary consisting of the following keys:

        rateLimit
            requests per second at steady rate, float

        burstLimit
            maximum number of requests per second, integer

    quota
        A dictionary consisting of the following keys:

        limit
            number of allowed requests per specified quota period [required if quota parameter is present]

        offset
            number of requests to be subtracted from limit at the beginning of the period [optional]

        period
            quota period, must be one of DAY, WEEK, or MONTH. [required if quota parameter is present

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.update_usage_plan plan_id='usage plan id' throttle='{"rateLimit": 10.0, "burstLimit": 10}'

    '''
    try:
        _validate_throttle(throttle)
        _validate_quota(quota)

        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

        patchOperations = []

        if throttle is None:
            patchOperations.append({'op': 'remove', 'path': '/throttle'})
        else:
            if 'rateLimit' in throttle:
                patchOperations.append({'op': 'replace', 'path': '/throttle/rateLimit', 'value': str(throttle['rateLimit'])})  # future lint: disable=blacklisted-function
            if 'burstLimit' in throttle:
                patchOperations.append({'op': 'replace', 'path': '/throttle/burstLimit', 'value': str(throttle['burstLimit'])})  # future lint: disable=blacklisted-function

        if quota is None:
            patchOperations.append({'op': 'remove', 'path': '/quota'})
        else:
            patchOperations.append({'op': 'replace', 'path': '/quota/period', 'value': str(quota['period'])})  # future lint: disable=blacklisted-function
            patchOperations.append({'op': 'replace', 'path': '/quota/limit', 'value': str(quota['limit'])})  # future lint: disable=blacklisted-function
            if 'offset' in quota:
                patchOperations.append({'op': 'replace', 'path': '/quota/offset', 'value': str(quota['offset'])})  # future lint: disable=blacklisted-function

        if patchOperations:
            res = conn.update_usage_plan(usagePlanId=plan_id,
                                         patchOperations=patchOperations)
            return {'updated': True, 'result': res}

        return {'updated': False}

    except ClientError as e:
        return {'error': __utils__['boto3.get_error'](e)}
    except (TypeError, ValueError) as e:
        return {'error': six.text_type(e)}


def delete_usage_plan(plan_id, region=None, key=None, keyid=None, profile=None):
    '''
    Deletes usage plan identified by plan_id

    .. versionadded:: 2017.7.0

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.delete_usage_plan plan_id='usage plan id'

    '''
    try:
        existing = describe_usage_plans(plan_id=plan_id, region=region, key=key, keyid=keyid, profile=profile)
        # don't attempt to delete the usage plan if it does not exist
        if 'error' in existing:
            return {'error': existing['error']}

        if 'plans' in existing and existing['plans']:
            conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
            res = conn.delete_usage_plan(usagePlanId=plan_id)
        return {'deleted': True, 'usagePlanId': plan_id}
    except ClientError as e:
        return {'error': __utils__['boto3.get_error'](e)}


def _update_usage_plan_apis(plan_id, apis, op, region=None, key=None, keyid=None, profile=None):
    '''
    Helper function that updates the usage plan identified by plan_id by adding or removing it to each of the stages, specified by apis parameter.

    apis
        a list of dictionaries, where each dictionary contains the following:

        apiId
            a string, which is the id of the created API in AWS ApiGateway

        stage
            a string, which is the stage that the created API is deployed to.

    op
        'add' or 'remove'
    '''
    try:
        patchOperations = []
        for api in apis:
            patchOperations.append({
                                    'op': op,
                                    'path': '/apiStages',
                                    'value': '{0}:{1}'.format(api['apiId'], api['stage'])
                                   })
        res = None
        if patchOperations:
            conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
            res = conn.update_usage_plan(usagePlanId=plan_id,
                                         patchOperations=patchOperations)
        return {'success': True, 'result': res}
    except ClientError as e:
        return {'error': __utils__['boto3.get_error'](e)}
    except Exception as e:
        return {'error': e}


def attach_usage_plan_to_apis(plan_id, apis, region=None, key=None, keyid=None, profile=None):
    '''
    Attaches given usage plan to each of the apis provided in a list of apiId and stage values

    .. versionadded:: 2017.7.0

    apis
        a list of dictionaries, where each dictionary contains the following:

        apiId
            a string, which is the id of the created API in AWS ApiGateway

        stage
            a string, which is the stage that the created API is deployed to.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.attach_usage_plan_to_apis plan_id='usage plan id' apis='[{"apiId": "some id 1", "stage": "some stage 1"}]'

    '''
    return _update_usage_plan_apis(plan_id, apis, 'add', region=region, key=key, keyid=keyid, profile=profile)


def detach_usage_plan_from_apis(plan_id, apis, region=None, key=None, keyid=None, profile=None):
    '''
    Detaches given usage plan from each of the apis provided in a list of apiId and stage value

    .. versionadded:: 2017.7.0

    apis
        a list of dictionaries, where each dictionary contains the following:

        apiId
            a string, which is the id of the created API in AWS ApiGateway

        stage
            a string, which is the stage that the created API is deployed to.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_apigateway.detach_usage_plan_to_apis plan_id='usage plan id' apis='[{"apiId": "some id 1", "stage": "some stage 1"}]'

    '''
    return _update_usage_plan_apis(plan_id, apis, 'remove', region=region, key=key, keyid=keyid, profile=profile)
